class Delivery < ActiveRecord::Base
  include MetadataContainer
  include UrlGeneration
  include ActionController::UrlWriter # for generating paths in notification emails
  extend ActiveSupport::Memoizable
  include Packages::TransferInfo
  include Packages::FileBrowsing

  attr_accessible :title, :note, :sources, :sender,
      :recipients, :private_recipients,
      :notifiable_on_upload, :notifiable_on_download, :notify_on_upload, :notify_on_download,
      :metadata, :delete_after_download_policy,
      :use_encryption_at_rest, :skip_metadata_validation, :skip_path_validation,
      :e_package, :disallow_dropbox, :forward, :ear_passphrase,
      :uuid, :e_package_id, :sender_id, :sender_name, :invitation,
      :recipient_string, :relay, :metadata_json, :skip_source_validation,
      :package_uuid, :error_code, :error_message, :skip_recipient_validation,
      :transfer_type, :warnings

  # legacy API option that we just ignore
  attr_accessible :send_upload_result

  belongs_to :e_package
  belongs_to :delivery_completion, foreign_key: :sequence_id

  has_many :user_deliveries
  has_many :workgroup_deliveries
  has_many :recipient_deliveries, :dependent => :destroy
  has_many :notifications, dependent: :destroy
  has_many :transfer_requests, :dependent => :destroy
  has_many :e_sessions
  has_many :upload_sessions,
    through: :e_package,
    class_name: 'ESession',
    conditions: ["direction = ? OR direction = ?", ESession::DIRECTION_UPLOAD, ESession::DIRECTION_CREATE]
  has_many :upload_requests,
    through: :e_package,
    class_name: 'TransferRequest',
    conditions: ["direction = ? OR direction = ?", ESession::DIRECTION_UPLOAD, ESession::DIRECTION_CREATE]

  has_many :package_relays, :conditions => ["relay_type = '0'"]
  has_one  :invitation

  belongs_to :sender, :class_name => 'User', :foreign_key => 'sender_id'
  named_scope :completed, :joins => :e_package, :conditions => ["e_packages.e_package_upload_status = ?", "Completed"],
    :order => "GREATEST(e_packages.e_package_upload_completion_time, deliveries.created_at) ASC"

  before_validation_on_create :make_package, :make_recipient_deliveries, :process_metadata
  before_save :generate_sequence_id, if: :forward?

  validate_on_create :check_prescence_of_recipients
  validate_on_create :check_recipients
  validate_on_create :check_notifications
  validate_on_create :check_sources, :unless => :skip_source_validation
  validate_on_create :check_metadata, :unless => :skip_metadata_validation
  validate_on_create :check_paths, :unless => :skip_path_validation
  validate_on_create :check_ear
  validates_length_of :title, :maximum => 245
  validates_length_of :note, :maximum => 8000
  validates_presence_of :title

  attr_accessor :use_encryption_at_rest, :sources, :metadata,
      :delete_after_download_policy, :skip_metadata_validation,
      :skip_path_validation, :disallow_dropbox,
      :invalid_recipient_names, :invalid_notifications, :ear_passphrase, :skip_source_validation,
      :package_uuid, :notify_on_upload, :notify_on_download, :error_code, :error_message, :skip_recipient_validation,
      :transfer_type, :warnings

  # legacy API option that we just ignore
  attr_accessor :send_upload_result

  validates_length_of :ear_passphrase, :minimum => 1, if: Proc.new { |d| !d.ear_passphrase.nil? && d.ear_passphrase.blank? }

  default_value_for(:uuid) {SecureRandom.uuid}
  default_value_for :use_encryption_at_rest, false
  default_value_for :skip_path_validation, false
  default_value_for :skip_source_validation, false
  default_value_for :sources, []

  default_value_for :notify_on_upload, false
  default_value_for :notify_on_download, false
  default_value_for :invalid_notifications, []

  default_value_for :invalid_recipient_names, []
  default_value_for :metadata, {}
  default_value_for :note, ""

  default_value_for :skip_recipient_validation, false
  default_value_for :warnings, {}

  after_save :save_metadata_to_pkg_dir, if: :save_metadata?

  after_save do |record|
    if record.forward? && !record.relay?
      record.e_package.deliver_upload_result(record)
    end
  end

  def after_initialize
    if sender
      self.sender_name ||= sender.name
      # this attribute is horribly named, it is also used for non-dropbox submissions from invites
      self.dbox_extusr_sender_pkg_passcode ||= SecureRandom.uuid
    end
  end

  def private_recipients=(str_or_arr)
    recipient_names = str_or_arr.is_a?(Array) ? str_or_arr : Utils.parse_recipients(str_or_arr)
    recipient_names = expand_distribution_lists(recipient_names).uniq
    recipient_names.each do |name|
      add_recipient(name, false)
    end
    str_or_arr
  end

  def public_recipients=(str_or_arr)
    recipient_names = str_or_arr.is_a?(Array) ? str_or_arr : Utils.parse_recipients(str_or_arr)
    recipient_names = expand_distribution_lists(recipient_names).uniq
    recipient_names.each do |name|
      add_recipient(name, true)
    end
    str_or_arr
  end
  alias_method :recipients=, :public_recipients=

  def add_recipient(name, public)
    self.invalid_recipient_names ||= []
    @recipient_hashes ||= {}
    recipient = lookup_recipient(name)
    if recipient
      @recipient_hashes[recipient] ||= {}
      @recipient_hashes[recipient][:name] = name
      @recipient_hashes[recipient][:public] ||= public
    else
      invalid_recipient_names << name
    end
  end

  def all_recipients
    recipient_deliveries.map(&:recipient_name)
  end

  def public_recipients
    recipient_deliveries.public.map(&:recipient_name)
  end
  alias_method :recipients, :public_recipients

  def private_recipients
    recipient_deliveries.private.map(&:recipient_name)
  end

  def http_transfer?
    transfer_type == 'http'
  end

private

  def expand_distribution_lists(recipient_names)
    expanded_list = []
    recipient_names.each do |name|
      recipient = lookup_recipient(name)
      if recipient.is_a? DistributionList
        expanded_list += expand_distribution_lists(recipient.contact_names)
      else
        expanded_list << name
      end
    end
    expanded_list
  end

  def make_package
    return if e_package
    self.e_package = EPackage.new(
      e_package_name:               title,
      e_package_sender_id:          sender.id,
      e_package_sender_name:        sender.name,
      e_package_note:               note,
      ear_encrypted:                ear_encrypted?,
      delete_after_download_policy: delete_after_download_policy,
      source_share_id:              sources.first && sources.first[:id],
      source_paths:                 sources.first && sources.first[:paths],
    )
    e_package.e_package_uuid = package_uuid if package_uuid
  end

  def ear_encrypted?
    require_ear? || ear_passphrase.present? || use_encryption_at_rest
  end

  def process_metadata
    self.metadata['_pkg_uuid'] = e_package.e_package_uuid
    self.metadata['_pkg_name'] = title
    self.metadata['_created_utc'] = Time.now.utc
    if dropbox_delivery?
      self.metadata['_dropbox_name'] = workgroup_deliveries.first.recipient_name
    end
    if metadata_profile
      self.metadata['_metadata_profile_name'] = metadata_profile.name
      self.metadata['_metadata_profile_uuid'] = metadata_profile.uuid
      self.metadata['_metadata_profile_version'] = metadata_profile.updated_at.utc.strftime('%Y-%m-%d %T')
    end
    self.metadata_json = metadata.to_json
  end

  def make_recipient_deliveries
    @recipient_hashes ||= {}
    @recipient_hashes.each_pair do |recipient, recipient_attributes|
      case recipient
      when EWorkgroup
        make_workgroup_delivery(recipient, recipient_attributes)
      when User
        make_user_delivery(recipient, recipient_attributes)
      else
        invalid_recipient_names << recipient_attributes[:name]
      end
    end
  end

  def make_workgroup_delivery(recipient, recipient_attributes)
    if recipient.dropbox? && disallow_dropbox
      invalid_recipient_names << recipient_attributes[:name]
    else
      workgroup_deliveries << WorkgroupDelivery.new(
        recipient:      recipient,
        recipient_name: recipient_attributes[:name],
        delivery:       self,
        e_package:      e_package,
        share:          recipient.destination_share || Share.default_destination,
        private:        !recipient_attributes[:public]
      )
    end
  end

  def make_user_delivery(recipient, recipient_attributes)
    user_deliveries << UserDelivery.new(
      recipient:      recipient,
      recipient_name: recipient_attributes[:name],
      delivery:       self,
      e_package:      e_package,
      share:          Share.default_destination,
      private:        !recipient_attributes[:public]
    )
  end

  def lookup_recipient(name)
    name = name.sub(/^\*/, '')
    sender.lookup_recipient(name)
  end

  def check_prescence_of_recipients
    errors.add(:base, "No valid recipients") if @recipient_hashes.empty?
  end

  def check_recipients
    invalid_recipient_names.each do |name|
      warnings[:invalid_recipients] ||= []
      !skip_recipient_validation ? errors.add(:base, "Invalid recipient: #{name}") : warnings[:invalid_recipients].push(name)
    end
  end

  def check_notifications
    if ActiveRecord::ConnectionAdapters::Column.value_to_boolean @notify_on_upload
      Utils.parse_recipients(notifiable_on_upload).each do |name|
        if valid_cc? name
          build_notification name, 'upload'
        else
          errors.add(:base, "Invalid cc (on upload): #{name}")
        end
      end
    end

    if ActiveRecord::ConnectionAdapters::Column.value_to_boolean @notify_on_download
      Utils.parse_recipients(notifiable_on_download).each do |name|
        if valid_cc? name
          build_notification name, 'download'
        else
          errors.add(:base, "Invalid cc (on download): #{name}")
        end
      end
    end
  end

  def check_metadata
    return unless metadata_profile.present?
    json_errors = metadata_profile.json_errors(metadata_json)
    json_errors.each { |e| errors.add(:metadata, e) }
  end

  def check_sources
    if sources.count != 1
      errors.add(:sources, 'Must have exactly one source')
    else
      source_share = SourceShare.find_by_id(sources.first[:id])
      if source_share.nil?
        errors.add(:sources, "Unknown source share id #{sources.first[:id]}")
      elsif !sender.accessible_shares.include?(source_share)
        errors.add(:sources, 'Source not authorized')
      elsif source_share.remote? && !sender.can_create_from_remote?
        errors.add(:sources, 'Source not authorized')
      end
    end
  end

  def check_paths
    source_paths = sources.first[:paths]
    unless source_paths && source_paths.count > 0
      errors.add_to_base("No items to upload!")
      return
    end
    basenames = source_paths.map {|sp| File.basename(sp) }
    unless basenames == basenames.uniq
      errors.add_to_base("Filenames must be unique")
    end
  end

  def check_ear
    if sources.present?
      share = SourceShare.find_by_id(sources.first[:id])
      if share && share.remote? && require_ear? && ear_passphrase.blank?
        errors.add(:ear_passphrase, 'must be specified')
      end
    end
    if disallow_ear? && (use_encryption_at_rest || ear_passphrase.present?)
      errors.add(:ear_passphrase, 'not allowed')
      self.error_code = ErrorCodes::EAR_NOT_ALLOWED
      self.error_message = "Encryption-at-rest is disabled, passphrase is not allowed."
    end
  end

  def disallow_ear?
    if dropbox_delivery?
      dropbox.disallow_ear?
    else
      EConfiguration.disallow_ear?
    end
  end

public

  def self.search_for(search, options = {})
    page = options.delete(:page) || 1
    paginate({
      page: page,
      joins: :recipient_deliveries,
      select: 'DISTINCT deliveries.*',
      conditions: [
        'sender_name LIKE :search OR
         title       LIKE :search OR
         note        LIKE :search OR
         recipient_deliveries.recipient_name LIKE :search',
        { search: "%#{search}%" }
      ]
    }.merge(options))
  end

  def self.paginated_sent_list(page, user, order)
    paginate(
      :page => page,
      :per_page => user.rows_per_page,
      :conditions => [
        "deliveries.sender_id=? AND " +
        "deliveries.archived=false AND " +
        "e_packages.e_package_status = ? AND " +
        "e_packages.e_package_files_deleted_by_admin_time IS NULL",
        user.id,
        EPackage::STATUS_VALID
      ],
      :order => order,
      :include => [:e_package, :sender]
    )
  end

  def self.paginated_sent_history(page, user, order)
    paginate(
      :page => page,
      :per_page => user.rows_per_page,
      :conditions => [
        "deliveries.sender_id=?", user.id
      ],
      :order => order,
      :include => [:e_package, :sender]
    )
  end

  def sender_email
    u = User.find(sender_id)
    u ? u.email : ''
  end

  def recipient_emails
    s = user_deliveries.collect{|d| d.recipient.email }.compact.uniq.join(',')
    s
  end

  # Assumption: sender is local
  def remote_servers
    this_server = HomeServer.this_server
    relay_servers = []
    each_recipient do |recipient|
      recipient.home_servers.each do |home_server|
        relay_servers << home_server if ((home_server != this_server) && !relay_servers.include?(home_server))
      end
    end
    relay_servers
  end

  def recipients_for_home_server(home_server)
    recipients = []
    each_recipient do |recipient|
      recipients << recipient if recipient.home_servers.include?(home_server)
    end
    recipients
  end

  def external_recipients_for_home_server(home_server)
    recipients = []
    each_recipient do |recipient|
      recipients << recipient if (recipient &&
                                  recipient.is_a?(User) &&
                                  recipient.external_email_user? &&
                                  (recipient.home_server == home_server))
    end
    recipients
  end

  def each_recipient
    self.all_recipients.each do |d|
      if d.first=='*'
        workgroup = EWorkgroup.find_by_e_wg_name(d[1,d.length-1])
        yield workgroup if workgroup
      else
        recipient = User.find_by_name(d)
        yield recipient if recipient
      end
    end
  end

  def remove_from_senders_view
    # remove the delivery only if the package is not uploading
    # (state might have changed since the web browser link was clicked)
    unless self.e_package.busy_uploading?
      self.archived = true
      unless self.save
        logger.error("Error trying to save changed removed_by_sender " +
                     "field for delivery id=" + self.id.to_s)
      end
    end
  end

  def send_received_notifications(share)
    notified_users = Set.new
    to_notify = user_deliveries.find_all_by_share_id(share.id) + workgroup_deliveries.find_all_by_share_id(share.id)
    to_notify.each do |recipient_delivery|
      notified_users += recipient_delivery.send_received_notifications(notified_users)
    end
  end

  def send_del_dled_notif(downloader, partial_package)
    package = EPackage.find self.e_package_id
    sender  = User.find self.sender_id

    if downloader.notify_downloader_on_download
      package_url = EConfiguration.get_full_main_server_address + received_path(RecipientDelivery.find_by_delivery_id(self.id), local: downloader.force_local_login?)
      BackgroundNotifier.deliver_package_downloaded_notification(downloader, package, self, package_url, partial_package, downloader)
    end

    package_url = EConfiguration.get_full_main_server_address + sent_path(self.id, local: sender.force_local_login?)

    return if package.full_complete_download_count_by_user(downloader) > 1
    return if package.partial_complete_download_count_by_user(downloader) > 1
    return if sender.name == downloader.name

    notifications.on_download.each do |n|
      if sender.email == n.email
        BackgroundNotifier.deliver_package_downloaded_notification(downloader, package,
                                                         self, package_url,
                                                         partial_package, sender)
      else
        BackgroundNotifier.deliver_package_downloaded_cc_notification(downloader, package, self, n)
      end
    end
  end

  def has_forwards?
    !Delivery.find_by_forward_and_e_package_id(true, e_package_id).nil?
  end

  def has_relays?
    !self.package_relays.empty?
  end

  def relay_status
    return nil if !has_relays?
    conditions = ["delivery_id = ?", self.id]
    state_min = PackageRelay.minimum(:relay_state, :conditions => conditions)
    state_min
  end

  def pretty_relay_status
    case relay_status
    when PackageRelay::PENDING
      "Relay Pending"
    when PackageRelay::RETRY
      "Retry Relay"
    when PackageRelay::RELAYING
      "Relaying"
    when PackageRelay::FAILED
      "Relay Failed"
    when PackageRelay::COMPLETE
      "Complete"
    end
  end

  def metadata_and_note
    data = note
    if !user_metadata_fields.empty?
      data = "#{self.metadata_lines}\nNote:\n #{data}"
    end
    data
  end

  def metadata_lines
    mdata = ""
    user_metadata_fields.each do |key|
      mdata += "#{key}: #{metadata_hash[key]}\n"
    end
    mdata
  end

  def metadata_xml(options = {})
    builder = options[:builder] || Builder::XmlMarkup.new(options)
    builder.metadata do
      metadata = ActiveSupport::JSON.decode(metadata_json)
      if metadata
        metadata.each_pair do |k, v|
          builder.field(v, :name => k)
        end
      end
    end
  end

  def save_metadata_to_pkg_dir
    share = default_destination

    save_path = metadata_file_save_path

    share.api_client.create_directory(save_path)

    builder = Builder::XmlMarkup.new(:indent => 2)
    builder.instruct!
    contents = metadata_xml(:builder => builder)

    path = File.join(save_path, 'aspera-metadata.xml')
    share.api_client.create_file(path, contents)
  rescue
    logger.warn("Failed to save metadata XML file for dropbox package #{e_package_id}")
    logger.debug($!.backtrace.join("\n"))
  end

  def metadata_file_save_path
    if FASPEX_CONFIG['SaveMetadataInPackage']
      full_package_path(default_destination)
    else
      outer_package_path(default_destination)
    end
  end

  def upload_from_source
    log_msg = "Uploading package (id: #{e_package_id}) from source share #{e_package.source_share.name}"
    logger.debug(log_msg)
    transfer_job = create_upload_job(make_symlinks_for_default_destination?, make_cloud_references?)
    transfer_job.perform
    if make_symlinks_for_default_destination? || make_cloud_references?
      create_fake_upload(e_package.source_share, default_destination)
    end
  rescue => e
    logger.warn("#{log_msg} failed")
    logger.debug(e)
    logger.debug(e.backtrace.join("\n"))
    raise EPackage::UploadFromSourceFailed, "#{log_msg} failed"
  end

  def node_path(relative_path)
    sanitized_path = Utils.expand_path(relative_path, '/')
    File.join(inner_directory_path, sanitized_path)
  end

  def node_paths(paths_in_pkg)
    paths_in_pkg.map {|path| node_path(path)}
  end


  def dropbox
    recipient = nil
    if workgroup_deliveries.length == 1
      recipient = workgroup_deliveries.first.recipient
    elsif @recipient_hashes && @recipient_hashes.length == 1
      recipient = @recipient_hashes.keys.first
    end

    if recipient.is_a?(EWorkgroup) && recipient.dropbox?
      recipient
    else
      nil
    end
  end
  memoize :dropbox

  def dropbox_delivery?
    invalid_recipient_names.blank? && dropbox.present?
  end

  def metadata_profile
    if dropbox_delivery?
      MetadataProfile.find_by_id(workgroup_deliveries.first.recipient.dbox_mdata_profile_id)
    else
      MetadataProfile.find_by_id(EConfiguration.first.normal_pkg_mdata_profile_id)
    end
  end
  memoize :metadata_profile

  def save_metadata?
    return false if forward?
    if dropbox_delivery?
      workgroup_deliveries.first.recipient.dbox_save_metadata_to_file?
    else
      EConfiguration.first.save_metadata_to_file?
    end
  end

  def delivery
    self
  end

  def completed_at
    if e_package.completed?
      [e_package.e_package_upload_completion_time, created_at].max
    else
      nil
    end
  end

  def generate_sequence_id
    self.delivery_completion = DeliveryCompletion.create # generate next sequence_id
  end

  def generate_sequence_id!
    generate_sequence_id
    save!
  end

  def storage_share
    if dropbox_delivery? && dropbox.destination_share && dropbox.upload_directly_to_custom_inbox?
      dropbox.destination_share
    else
      default_destination
    end
  end

  def destination_shares
    share_ids = recipient_deliveries.map(&:share_id).uniq - [default_destination(true).id]
    shares = share_ids.map {|share_id| Share.find_by_id(share_id)}
    shares.reject(&:nil?)
  end

  def relay_to_inboxes
    destination_shares.each {|dest_share| relay_to_inbox(dest_share)}
  end

  def full_package_path(share)
    File.join(share.directory, File.join(e_package.outer_directory_name, e_package.inner_directory_name))
  end

  def outer_package_path(share)
    File.join(share.directory, e_package.outer_directory_name)
  end

  # Only skip inbox if workgroup is sole recipient
  def workgroup_skip_inbox?
    if recipients.length == 1
      recipient = lookup_recipient(recipients.first)
      if recipient.is_a?(EWorkgroup) && recipient.upload_directly_to_custom_inbox
        return true
      end
    end
  end

  # Determine if there is a custom inbox and workgroup is only recipient
  def workgroup_with_custom_inbox?
    if recipients.length == 1
      recipient = lookup_recipient(recipients.first)
      if recipient.is_a?(EWorkgroup)
        return true
      end
    end
  end

  def workgroup_recipient
    workgroup_with_custom_inbox? ? lookup_recipient(recipients.first) : nil
  end

  def relay_to_inbox(dest_share, force = false)
    recipient_delivery = recipient_deliveries.first(conditions: {share_id: dest_share.id})
    if recipient_delivery.inner_directory_exists? && !force
      send_received_notifications(dest_share)
    else
      default_destination_share = default_destination
      direction = ESession::DIRECTION_RELAY
      source_node = default_destination_share.node
      source_paths = [outer_package_path(default_destination_share)]
      dest_path = dest_share.directory
      cookie = recipient_delivery.prepare_transfer(direction, ['/'], sender).cookie

      ### If the source node is the same as the dest node, and symlinking is enabled,
      ### make symlink instead of transferring from default inbox back to custom inbox on source node
      symlink = make_symlinks_for_custom_inbox?(e_package.source_share, dest_share)

      if symlink
        # Change source from default inbox to original source
        source_node = e_package.source_share.node
        source_paths = e_package.source_share.node_paths(e_package.e_package_paths.map(&:e_uploader_local_path))
        # Must build full package path because we're not copying package dir from default inbox
        dest_path = full_package_path(dest_share)
      end

      transfer_params = TransferParams.new(
          cookie: cookie,
          target_rate_kbps: select_relay_bandwidth(default_destination_share, dest_share),
          paths: [{destination: dest_path}]
      )

      TransferJob.new(source_node, source_paths, dest_share.node, dest_path, transfer_params, symlink: symlink).perform

      create_fake_upload(default_destination_share, dest_share, ESession::DIRECTION_RELAY) if symlink
    end
  end

  def relay_bandwidth dest_share
    select_relay_bandwidth(default_destination, dest_share)
  end

  def forward_template
    result = "\n\n"
    if user_metadata_fields.any?
      result += "--------- Forwarded metadata ---------\n"
      result += delivery.user_metadata_fields.map { |key| "#{key}: #{delivery.metadata_hash[key]}" }.join("\n")
      result += "\n"
    end
    result += "--------- Forwarded note ---------\n"
    result += note
    result
  end

  def make_cloud_references?
    source_share = SourceShare.find_by_id(sources.first[:id])
    source_share_node = source_share.node
    default_destination_node = default_destination.node

    # If cloud referencing feature is disabled in the app then
    # reset enable_cloud_referencing to false in case it's already turned ON.
    # TODO: Remove the following line when ES 3.6 is available to support cloud referencing.
    source_share.update_attribute(:enable_cloud_referencing, false) if !App.cloud_referencing_feature_enabled?

    @make_cloud_references ||= source_share.uses_azure_storage? &&
      source_share.enable_cloud_referencing? &&
      ( source_share_node == default_destination_node ||
        source_share_node.access_key == default_destination_node.access_key )
  end

  def pkg_contains_directories?
    source_share = SourceShare.find_by_id(sources.first[:id])
    source_paths = source_share.node_paths(sources.map{ |s| s['paths']}.flatten)
    source_paths.any? do  |source_path|
      source_share.node.api_client.files_info(source_path).response_parsed['paths'][0]['type'] == 'directory'
    end
  end

  # If a package has multiple deliveries, then all after the first are forwards - use original delivery for share / path info.
  def original_delivery
    e_package.deliveries.first rescue self
  end

  def is_original_delivery?
    self == original_delivery
  end

  def default_destination(ignore_original = false)
    if is_original_delivery? || ignore_original
      workgroup_skip_inbox? ? workgroup_recipient.destination_share : Share.default_destination
    else
      original_delivery.default_destination
    end
  end

  def default_download_location(ignore_original = false)
    download_share = Share.default_destination
    if is_original_delivery? || ignore_original
      download_share = workgroup_recipient.destination_share if workgroup_with_custom_inbox?
    else
      download_share = original_delivery.default_download_location
    end

    download_share || default_destination(ignore_original)
  end

  def default_destination_path(ignore_original = false)
    if is_original_delivery? || ignore_original
      workgroup_skip_inbox? ? full_package_path(workgroup_recipient.destination_share) : e_package.node_path('/')
    else
      original_delivery.default_destination_path
    end
  end

  def default_transfer_host(request, upload=false)
    transfer_share = upload ? default_destination : default_download_location
    transfer_share.node.get_server_address(request.remote_ip, request.host)
  end

  def create_fake_upload(source, dest, direction = ESession::DIRECTION_CREATE)
    share = dest
    stats = default_destination.node.api_client.du(full_package_path(default_destination)).response_parsed
    total_size = stats['total_size']
    file_count = stats['file_count']
    now = Time.now

    transfer_request = TransferRequest.find_by_delivery_id_and_share_id(self.id, share.id)

    sess = ESession.create(
        e_package: e_package,
        status: ESession::STATUS_COMPLETE,
        initiator: e_package.sender,
        direction: direction,
        delivery: e_package.deliveries.first,
        central_id: SecureRandom.uuid,
        system_user_name: '',
        remote_host: source.node.host,
        error_code: 0,
        files_completed: file_count,
        files_failed: 0,
        bytes_pretransfer: total_size,
        bytes_written: total_size,
        bytes_transferred: total_size,
        share: share,
        e_session_start_notification_time: now,
        e_session_stop_notification_time: now,
        transfer_request: transfer_request
    )

    e_package.update_from_session(sess)
  end

private

  def valid_cc? name
    Utils.valid_email?(name) || sender.lookup_user_recipient(name)
  end

  def build_notification name, event
    if user = AuthorizableUser.find_by_name(name)
      notifications.build name: user.full_name, email: user.email, event: event
    else
      notifications.build name: name, email: name, event: event
    end
  end

  def make_symlinks_for_custom_inbox?(source, dest)
    @make_symlinks ||= source.enable_linking &&
        source.node == dest.node &&
        ear_passphrase.nil? &&
        source.node.supports_symlinks?
  end

  def make_symlinks_for_default_destination?
    @make_symlinks ||= e_package.source_share.enable_linking? &&
        e_package.source_share.node == default_destination.node &&
        ear_passphrase.nil? &&
        e_package.source_share.node.supports_symlinks?
  end

  def create_upload_job(symlink = false, cloud_reference = false)
    source_paths = e_package.source_share.node_paths(e_package.e_package_paths.map(&:e_uploader_local_path))
    cookie = prepare_transfer(ESession::DIRECTION_CREATE, source_paths, sender).cookie
    transfer_params = TransferParams.new(
        cookie: cookie,
        target_rate_kbps: select_relay_bandwidth(e_package.source_share, default_destination),
        paths: [{destination: default_destination_path}]
    )
    TransferJob.new(
        e_package.source_share.node,
        source_paths,
        default_destination.node,
        default_destination_path,
        transfer_params,
        ear_passphrase: ear_passphrase,
        symlink: symlink,
        cloud_reference: cloud_reference
    )
  end

  def select_relay_bandwidth(source_share, destination_share)
    if source_share.override_transfer_parameters? && destination_share.override_transfer_parameters?
      [source_share.relay_outgoing_bandwidth, destination_share.relay_incoming_bandwidth].min
    elsif source_share.override_transfer_parameters?
      [CConfiguration.first.c_relay_outgoing_bandwidth, source_share.relay_outgoing_bandwidth].min
    elsif destination_share.override_transfer_parameters?
      [CConfiguration.first.c_relay_outgoing_bandwidth, destination_share.relay_incoming_bandwidth].min
    else
      CConfiguration.first.c_relay_outgoing_bandwidth
    end
  end
end
