require 'net/http'
require 'utils'
require 'packages/package_methods'

class EPackage < ActiveRecord::Base
  class UploadFromSourceFailed < StandardError; end
  class SendError < StandardError; end
  class DeletionError < StandardError; end

  extend ActiveSupport::Memoizable
  include ActionController::UrlWriter
  include MetadataContainer
  include Packages::TransferInfo
  include Packages::FileBrowsing

  # package after-download expiration policies
  DO_NOTHING                           = 0
  DELETE_AFTER_ALL_RECIPIENTS_DOWNLOAD = 1
  DELETE_AFTER_FIRST_FULL_DOWNLOAD     = 2


  # package status states
  STATUS_VALID          = "Available"
  STATUS_DELETED_SENDER = "Cancelled by Sender"
  STATUS_DELETED_ADMIN  = "Deleted by Administrator"
  STATUS_DELETED_POLICY = "Deleted by Policy"
  CONTENTS_DELETED      = "Contents deleted"
  PENDING_DELETION      = "Pending deletion"

  FORWARDED_PACKAGE_PREFIX = ''

  belongs_to :e_package_sender, :class_name => "User", :foreign_key => "e_package_sender_id"
  belongs_to :source_share

  has_many :deliveries,       :dependent => :destroy, :order => :created_at
  has_many :recipient_deliveries, :through => :deliveries
  has_many :user_deliveries,  :dependent => :destroy
  has_many :workgroup_deliveries, :dependent => :destroy
  has_many :e_package_accesses,     :dependent => :destroy
  has_many :e_package_paths,        :dependent => :destroy
  has_many :e_sessions,             :dependent => :destroy
  has_many :transfer_requests, :dependent => :destroy

  has_many :completed_downloads, :class_name => 'ESession',
                                 :conditions => ['status=? AND direction=?',
                                                 ESession::STATUS_COMPLETE, ESession::DIRECTION_DOWNLOAD]

  has_many :completed_full_downloads, :class_name => 'ESession',
                                      :conditions => ['status=? AND direction=? AND partial_package=?',
                                                      ESession::STATUS_COMPLETE, ESession::DIRECTION_DOWNLOAD, false]

  has_many :completed_partial_downloads, :class_name => 'ESession',
                                         :conditions => ['status=? AND direction=? AND partial_package=?',
                                                         ESession::STATUS_COMPLETE, ESession::DIRECTION_DOWNLOAD, true]

  has_many :active_downloads, :class_name => 'ESession',
                              :conditions => ['status=? AND direction=?',
                                              ESession::STATUS_TRANSFERRING, ESession::DIRECTION_DOWNLOAD]

  has_many :upload_sessions,
    class_name: 'ESession',
    conditions: ["direction = ? OR direction = ?", ESession::DIRECTION_UPLOAD, ESession::DIRECTION_CREATE]
  has_many :upload_requests,
    class_name: 'TransferRequest',
    conditions: ["direction = ? OR direction = ?", ESession::DIRECTION_UPLOAD, ESession::DIRECTION_CREATE]

  validates_length_of :e_package_name, :maximum => 245
  validates_length_of :e_package_note, :maximum => 8000

  validates_presence_of :e_package_sender_id, :e_package_name

  attr_accessor :source_paths

  attr_accessible :e_package_name, :e_package_sender_id,
    :e_package_note, :ear_encrypted, :delete_after_download_policy,
    :source_share_id, :source_paths,
    :e_package_sender_name

  default_value_for(:e_package_server_hostname) {EConfiguration.get_server_address}
  default_value_for(:e_package_uuid) {SecureRandom.uuid}
  default_value_for :e_package_total_size_bytes,      0
  default_value_for :e_package_total_files,           0
  default_value_for :e_package_total_elapsed_seconds, 0
  default_value_for :e_package_status,                EPackage::STATUS_VALID
  default_value_for :e_package_upload_status,         ESession::STATUS_SUBMITTED

  before_validation_on_create :set_sender_name, :set_package_paths, :set_metadata_json
  before_save :set_outer_directory_name

  named_scope :missing, conditions: ['missing_since IS NOT NULL AND e_package_contents_exist = ?', true]
  named_scope :pending_deletion, lambda {
    { conditions: ['e_package_status = ? AND pending_deletion_time < ?', PENDING_DELETION, Time.now] }
  }

private

  def set_sender_name
    if sender
      self.e_package_sender_name ||=
          sender.external_email_user? ? "#{sender.name} (outside email user)" : sender.name
    end
  end

  def set_package_paths
    if source_paths.present?
      source_paths.each do |path|
        self.e_package_paths << EPackagePath.new(:e_uploader_local_path => path)
      end
    end
  end

  # open up the metadata_json to insert some values
  def set_metadata_json
    decoded_json = ActiveSupport::JSON.decode(metadata_json.to_s) || {}
    decoded_json['_pkg_uuid'] = e_package_uuid
    decoded_json['_created_utc'] = Time.now.utc
    self.metadata_json = decoded_json.to_json
  end

  def set_outer_directory_name
    # keeps outer directory name to 255 characters or less, UUID is 36 characters long
    self.outer_directory_name ||= "#{filesafe_title[0,201]} - #{e_package_uuid}.aspera-package"
  end

  #===============================================================================================
  # CLASS METHODS
  #===============================================================================================

public

  def self.get_expired_packages(max_age_in_days)
    earliest_valid_time = Time.now - (max_age_in_days * 1.day)
    EPackage.all(
      :order      => "created_at DESC",
      :conditions => [
        "(e_package_upload_completion_time < ? OR (created_at < ? AND e_package_upload_status != ?)) AND e_package_upload_status != ? AND e_package_status = ?",
        earliest_valid_time,
        earliest_valid_time,
        ESession::STATUS_COMPLETE,
        ESession::STATUS_TRANSFERRING,
        STATUS_VALID
      ]
    )
  end

  def self.deliver_failed_uploads
    EPackage.all(:conditions => {:e_package_upload_status => ESession::STATUS_SUBMITTED}).each do |package|
      if package.upload_timed_out?
        delivery = Delivery.find_by_e_package_id(package.id)
        package.deliver_upload_result(delivery)
        package.update_attribute(:e_package_upload_status, ESession::STATUS_TIMED_OUT)
      end
    end
  end

  def self.mark_expired_packages
    if EConfiguration.packages_expire?
      Utils.log_info("EPackage.mark_expired_packages...")
      Utils.log_debug("Marking expired packages for deletion...")

      get_expired_packages(EConfiguration.get_days_until_packages_expire).each do |p|
        if p.busy_uploading?
          Utils.log_debug("Package still uploading: #{p.id}:#{p.e_package_uuid}:#{p.e_package_name}")
        else
          p.mark_for_deletion
          Utils.log_debug("Package marked for deletion: #{p.id}:#{p.e_package_uuid}:#{p.e_package_name}")
        end
      end
      Utils.log_debug("Finished marking expired packages for deletion...")
    end
  end

  def self.clean_up_missing_packages
    # try to reattach missing packages
    missing.each do |p|
      p.browse_inner_directory # clears missing_since if contents now exist
    end

    missing.all(conditions: ['missing_since < ?', 1.day.ago]).each do |p|
      p.mark_as_deleted
    end
  end

  def self.delete_marked_packages
    pending_deletion.each do |p|
      begin
        p.delete_contents # raises error on deletion failure
      rescue
        # on deletion failure, we'll try again later
        p.pending_deletion_time = Time.now + RbConfig.delete_retry_period
        p.save
      end
    end
  end

  #===============================================================================================
  # INSTANCE METHODS
  #===============================================================================================

  def create_relays_for(delivery)
    Utils.log_debug("Creating relay records for package #{delivery.title.inspect}", :pkg_relay)
    delivery.remote_servers.each do |remote_home_server|
      PackageRelay.for(self, delivery, remote_home_server)
    end
  end

  def contents_available?
    e_package_contents_exist? && inner_directory_exists?
  rescue
    false
  end

  #copied from application_helper.rb/gr_pkg_file_status
  def completed_or_partial?
    if e_package_files_deleted_by_admin_time
      return false
    elsif !inner_directory_exists?
      return false
    elsif e_package_contents_exist?
      return completed?
    end
    false
  end

  def accessible_to_user(user_id, allow_normal_admins = false)
    begin
      _user = User.find(user_id)
    rescue
      return false #if user not found, don't crash
    end

    #senders have access to their own packages
    if self.e_package_sender_id == user_id
      return true
    end

    #check if the user personally received the package
    user_deliveries.each do |one_personal_delivery|
      if one_personal_delivery.recipient_id == user_id
        return true
      end
    end

    #check if one of the user's workgroups received the package
    return true if (self.recipient_workgroups & _user.e_workgroups).size > 0

    # check if user is an admin TODO need to change for multiple admins
    return true if _user.admin_with_privilege?(User::PERM_SUPER)
    return true if (_user.admin? && allow_normal_admins)
    false
  end

  def can_be_downloaded_by?(user, ip)
    if !user.downloads_allowed
      log_violation(user, "user is not permitted to download packages")
    elsif !user.download_from_ip_allowed?(ip)
      log_violation(user, "user is not permitted to download from ip #{ip}")
    elsif !self.accessible_to_user(user.id)
      log_violation(user, "user is not authorized to download this package")
    else
      return true
    end
    return false
  end

  def can_be_uploaded_by?(user, ip)
    if !user.uploads_allowed && e_workgroups
      log_violation(user, "user is not permitted to upload packages")
    elsif !user.upload_from_ip_allowed?(ip)
      log_violation(user, "user is not permitted to upload from ip #{ip}")
    elsif !self.sent_by?(user)
      log_violation(user, "user tried to upload, but is not the sender of the package")
    elsif self.e_package_upload_status == "Completed"
      log_violation(user, "cannot upload to a completed package")
    else
      return true
    end
    false
  end

  def can_be_forwarded_by?(user)
    if !user.forwarding_allowed
      log_violation(user, "user is not permitted to forward packages")
    elsif !self.accessible_to_user(user.id)
      log_violation(user, "user is not authorized to access package")
    else
      return true
    end
    false
  end

  def sent_by?(user)
    self.e_package_sender_id == user.id
  end

  def age_in_days
    ((Time.now - (self.e_package_upload_completion_time || self.e_package_upload_start_time || self.created_at)) / 86400).to_i
  end

  # handle callback
  def before_create
    self.e_package_uuid = SecureRandom.uuid if self.e_package_uuid.blank?
  end

  def busy_downloading?
    active_downloads.each do |session|
      #make sure after updating stats that the session is still active
      if [ESession::STATUS_TRANSFERRING,
          ESession::STATUS_SUBMITTED].include?(session.status)
        return true
      end
    end
    false
  end

  def upload_timed_out?
    upload_sessions.empty? && (Time.now - self.updated_at) > FASPEX_LIVE_UPDATE_TIMEOUT
  end

  # Delete directory containing package contents
  # true if successful
  # false if not, with error message in errors object
  # NOTE: generally will not report if unable to delete files (same as rm -rf)
  def delete_contents
    response = containing_node.api_client.delete([package_directory_path])
    if response.error? && (response.path.nil? || response.path['error']['code'] != 404) # any error except "not found"
      self.errors.add_to_base("Error deleting package contents")
      logger.error("error deleting package contents for package id #{id}: " +
          response.user_message)
      raise RuntimeError, response.user_message
    else
      mark_as_deleted
    end
  end

  def package_directory_path
    containing_share.node_path(outer_directory_name)
  end

  def mark_as_deleted
    self.e_package_files_deleted_by_admin_time = Time.now
    self.e_package_contents_exist = false
    self.e_package_total_size_bytes = 0
    self.e_package_total_files = 0
    self.e_package_status = CONTENTS_DELETED
    self.pending_deletion_time = nil
    unless self.save
      Utils.log_model_error("EPackage:mark_as_deleted: failed to save package", self)
      raise RuntimeError, "failed to save package after deleting"
    end
  end

  def mark_for_deletion
    self.e_package_status = PENDING_DELETION
    self.pending_deletion_time = Time.now
    save!
  end

  # raises DeletionError if failed to mark for deletion on remote server
  def mark_for_deletion_on_all_servers
    HomeServer.other_servers.each do |hs|
      mark_for_deletion_on_remote(hs)
    end
    mark_for_deletion
  end

  def mark_for_deletion_on_remote(home_server)
    path = mark_for_deletion_remote_admin_package_path(skip_relative_url_root: true, id: e_package_uuid)
    success, _ = MultiServer::RestApiSupport.call_remote_admin_api(home_server, path)
    unless success
      raise DeletionError, "Delete package with uuid #{e_package_uuid} failed on remote server #{home_server.name}"
    end
  end

  # override accessor to protect from nil values from db
  # TODO need this?
  def e_package_note
    self['e_package_note'] || ""
  end

  def filesafe_title
    #take package title, replace illegal characters with underscores
    #illegal chars are | < > " ? * : / \ $ as well nonprinting chars like tabs
    self.e_package_name.gsub(/[[:cntrl:]\|<>"\?\[\]\*:\/\\$`]/,'_').squeeze(" ").strip.gsub(/\.+\z/, "")
  end

  def get_active_download_stats
    active_download_stats = []

    self.active_downloads.each do |session|
      #make sure after updating stats that the session is still active
      if session.status == ESession::STATUS_TRANSFERRING
        active_download_stats <<
          {
            :username                             => User.find(session.initiator_user_id).name,
            :download_pct_complete                => session.progress_percentage,
            :download_estimated_seconds_remaining => session.seconds_remaining
          }
      end
    end
    active_download_stats
  end

  def get_user_downloads_hash
    _hash = {}
    _hash.default=0
    self.completed_downloads.each do |one_completed_download_session|
      if one_completed_download_session.partial_package.blank?
        _hash[one_completed_download_session.initiator_user_id] += 1
      end
    end
    _hash
  end

  def verbose_completed_downloads
    (self.completed_downloads.collect{ |session|
       { :id => session.id,
         :user_name => (User.find(session.initiator_user_id).name rescue '(deleted user)'),
         :start_time => session.e_session_start_notification_time,
         :stop_time => session.e_session_stop_notification_time,
         :seconds_elapsed => session.seconds_elapsed,
         :remote_ip => session.remote_host,
         :partial_package => session.partial_package
       }
     } +
     self.active_downloads.collect{ |session|
       { :id => session.id,
         :user_name => (User.find(session.initiator_user_id).name rescue '(deleted user)'),
         :start_time => session.e_session_start_notification_time,
         :stop_time => nil,
         :seconds_elapsed => session.seconds_elapsed,
         :remote_ip => session.remote_host,
         :partial_package => session.partial_package
       }
     }).sort_by {|session| session[:id]}
  end

  def get_completed_downloads_by_user
    completed_download_stats = []
    for key, value in get_user_downloads_hash
      begin
        _user_name = User.find(key).name
      rescue
        _user_name = "(deleted user)"
      end
      completed_download_stats << {:username => _user_name, :download_count => value}
    end
    completed_download_stats
  end

  def get_total_completed_downloads
    self.get_completed_downloads_by_user.inject(0) {|sum, x| sum + x[:download_count]}
  end

  def get_total_completed_partial_downloads
    self.get_completed_partial_downloads_by_user.inject(0) {|sum, x| sum + x[:partial_download_count]}
  end

  def get_completed_partial_downloads_by_user
    completed_partial_download_stats = []
    summary_hash = {}
    summary_hash.default=0
    #check if the download was part of a package
    self.completed_downloads.each do |one_completed_download_session|
      unless one_completed_download_session.partial_package.blank?
        summary_hash[one_completed_download_session.initiator_user_id] += 1
      end
    end

    for key, value in summary_hash
      begin
        _user_name = User.find(key).name
      rescue
        _user_name = "(deleted user)"
      end
      completed_partial_download_stats << {:username => _user_name, :partial_download_count => value}
    end
    return completed_partial_download_stats
  end

  ##
  #@todo - et - just pass the object, not the id
  def mark_as_downloaded_for_user(user_id)
    access_record = self.e_package_accesses.first(:conditions=>["user_id=?", user_id])
    unless access_record
      #create
      unless self.e_package_accesses.create(:user_id           => user_id,
                                            :e_package_id         => self.id,
                                            :e_package_downloaded => true,
                                            :e_package_viewed     => true) #if the user downloaded it, mark as effectively viewed

        logger.error('Error trying to mark package as downloaded: ' + self.errors.full_messages.join("\n"))
      else
        _arr = get_completed_downloads_by_user.collect{ |d| d[:username] }
      end
    else
      access_record.e_package_downloaded = true
      unless access_record.save
        logger.error('Error trying to mark package as downloaded: ' + access_record.errors.full_messages.join("\n"))
      end
    end
  end

  def mark_as_viewed_for_user(user_id)
    access_record = self.e_package_accesses.find_or_create_by_user_id_and_e_package_id(user_id, self.id)
    access_record.e_package_viewed = true
    unless access_record.save
      logger.error('Error trying to mark package as viewed: ' + access_record.errors.full_messages.join("\n"))
    end
  end

  def recipients
    recipient_deliveries.map(&:recipient)
  end

  def recipient_users
    user_deliveries.map(&:recipient)
  end

  def recipient_workgroups
    workgroup_deliveries.map(&:recipient)
  end

  def sender
    User.find(:first, :conditions => ['id=?', self.e_package_sender_id])
  end

  def default_download_location
    deliveries.first.default_download_location rescue storage_share
  end

  def storage_share
    if dropbox_package?
      dropbox = recipient_workgroups.first
    end

    if dropbox && dropbox.destination_share && dropbox.upload_directly_to_custom_inbox?
      dropbox.destination_share
    else
      deliveries.any? ? deliveries.first.default_destination : Share.default_destination
    end
  end

  def containing_share
    transfer_request = TransferRequest.find_by_e_package_id(id)
    transfer_request.nil? ? Share.default_destination : transfer_request.share
  end

  def storage_node
    storage_share.node
  end

  def containing_node
    containing_share.node
  end

  # breadth-first recursive calculation of the size of the package by querying
  # the storage node
  def calculate_size(path_in_pkg)
    items = items_at(path_in_pkg)
    items.inject(0) do |sum, item|
      case item
      when DirectoryItem::File
        sum + item.size
      when DirectoryItem::Directory
        sum + calculate_size(package_path(item.path))
      else
        raise "Unknown item type"
      end
    end
  end

  # Checks the stats on the disk and corrects any errors in package size.
  def check_stats
    self.e_package_total_size_bytes = calculate_size('/')
    save!
  end

  def update_from_download_session(sess)
    #don't do anything, we don't explicitly handle download errors, just let them get filtered out of the status display in the
    # single package view by virtue of the fact that their sessions are no longer active
    return if [ESession::STATUS_STOPPED, ESession::STATUS_ERROR].include?(sess.status)

    # record successful downloads
    downloader = User.find(sess.initiator_user_id)
    self.mark_as_downloaded_for_user(downloader.id)

    # get TransferRequest from SOAP session object to determine if a complete package was downloaded
    delivery = Delivery.find(sess.delivery_id)

    if sess.status == ESession::STATUS_COMPLETE
      delivery.send_del_dled_notif(downloader, sess.transfer_request.partial_package?)

      if HomeServer.multi_server?
        # Asynchronous operation
        # MultiServer::RelayReceiver deletes packages from receiver side.
        PackageRelay.for_access(sess, delivery, downloader, sess.transfer_request.partial_package?)
      end

      # deletes locally
      delete_if_download_condition_met
    end
  end

  def full_complete_download_count_by_user(user)
    completed_full_downloads.
      all(:conditions => "initiator_user_id = #{user.id}").count
  end

  def partial_complete_download_count_by_user(user)
    completed_partial_downloads.
      all(:conditions => "initiator_user_id = #{user.id}").count
  end

  def has_been_fully_downloaded?
    users_with_access.each do |u|
      return true if has_been_fully_downloaded_by?(u)
    end
    false
  end

  def has_been_fully_downloaded_by?(user)
    !completed_full_downloads.all(:conditions => "initiator_user_id=#{user.id}").empty?
  end

  def has_been_fully_downloaded_by_all?
    users_with_access.each do |u|
      return false unless has_been_fully_downloaded_by?(u)
    end
    true
  end

  # return full list of all people that can download package
  def users_with_access
    users = user_deliveries.collect(&:recipient).select(&:active?)
    recipient_workgroups.each do |wg|
      memberships = wg.e_memberships.select(&:can_read)
      users += memberships.map(&:user).select(&:active?)
    end
    users.uniq
  end

  def delete_if_download_condition_met
    if delete_after_download?
      case delete_after_download_condition
      when DELETE_AFTER_FIRST_FULL_DOWNLOAD
        if has_been_fully_downloaded?
          mark_for_deletion
        end
      when DELETE_AFTER_ALL_RECIPIENTS_DOWNLOAD
        if has_been_fully_downloaded_by_all?
          mark_for_deletion
        end
      end
    end
  end

  def update_from_session(e_session)
    raise "ERROR: Attempt to update package from an unrelated session" if e_session.e_package != self

    case e_session.direction
    when ESession::DIRECTION_DOWNLOAD
      update_from_download_session(e_session)
    when ESession::DIRECTION_UPLOAD, ESession::DIRECTION_CREATE
      update_from_upload_session(e_session)
    when ESession::DIRECTION_RELAY
      update_from_relay_session(e_session)
    end
  end

  def update_from_relay_session(sess)
    if sess.status == ESession::STATUS_COMPLETE
      if sess.transfer_request.flatten
        send_relay_notifications(sess, :finish)
      else
        sess.delivery.send_received_notifications(sess.share)
      end
    elsif sess.transfer_request.flatten && (sess.status == ESession::STATUS_TIMED_OUT || sess.status == ESession::STATUS_ERROR)
      send_relay_notifications(sess, :error)
    end
  end

  def update_from_upload_session(sess)
    # update package stats from session data
    self.e_package_total_size_bytes = sess.bytes_total
    self.e_package_total_files      = sess.file_count
    self.update_package_status(sess)
    unless self.save
      logger.error('error while trying to save package with updated statistics: ' + package.errors.full_messages.join("\n"))
    end
    if sess.ended?
      # check_stats # disabled for S3 delay
      handle_upload_end
    end
  end

  def handle_upload_end
    delivery = Delivery.find_by_e_package_id(self.id)
    deliver_upload_result(delivery) unless delivery.relay?
    if completed?
      if invitation = delivery.invitation
        handle_external_submission_end invitation
      elsif self.dropbox_package?
        handle_dropbox_upload_end(delivery)
      end
      PostProcess.match_and_execute(delivery)
      delivery.generate_sequence_id!
      delivery.relay_to_inboxes unless delivery.relay?
      delivery.send_received_notifications(delivery.default_destination)
      create_relays_for(delivery) unless delivery.relay?
      relay_to_forwarding_destinations(delivery)
    end
  end

  def relay_to_forwarding_destinations(delivery)
    delivery.recipient_deliveries.find_all { |rd| rd.recipient_type == 'EWorkgroup' }.map(&:recipient).uniq.each do |e_wg|
      if e_wg.forwarding_enabled
        e_wg.e_workgroup_forwardings.each do |ewf|
          if ewf.forward
            dest_share = ewf.share
            relay_to_forwarding_destination(dest_share, delivery, e_wg, ewf)
            Utils.parse_recipients(ewf.notify_on_start_names || '').each do |name|
              notified_user = User.find_by_name(name)
              next if notified_user && !notified_user.active?
              BackgroundNotifier.deliver_workgroup_file_relay_notification(delivery, self, e_wg, name, dest_share.node.host, dest_share.directory, file_list_first_10, file_list_all, :start)
            end
          end
        end
      end
    end
  end

  def make_symlinks_for_relay_inbox?(source, dest)
    @make_symlinks ||= source.enable_linking &&
        source.node == dest.node &&
        !ear_encrypted? &&
        source.node.supports_symlinks?
  end

  def relay_to_forwarding_destination(dest_share, delivery, e_wg, ewf)
    recipient_delivery = delivery.recipient_deliveries.detect { |rd| rd.recipient == e_wg }
    overwrite_val = ewf.overwrite? ? 'Always' : 'Never'
    source_node = delivery.default_destination.node
    source_paths = [self.outer_directory_path]
    source_base = source_node.uses_azure_storage? ? URI.encode(source_node.api_client.destination_root + inner_directory_path) : inner_directory_path

    symlink = make_symlinks_for_relay_inbox?(source_share, dest_share)

    if symlink
      # Source is original files on this node
      source_node = source_share.node
      source_paths = source_share.node_paths(delivery.e_package.e_package_paths.map(&:e_uploader_local_path))
      #source_base = source_share.directory
      source_base = '/'
    end

    cookie = recipient_delivery.prepare_transfer(ESession::DIRECTION_RELAY, ['/'], delivery.sender, :share => dest_share, :flatten => true, :e_workgroup_id => e_wg.id).cookie
    transfer_params = TransferParams.new(
      cookie: cookie,
      target_rate_kbps: delivery.relay_bandwidth(dest_share),
      paths: [{destination: dest_share.directory}]
    )
    begin
      TransferJob.new(source_node, source_paths, dest_share.node, dest_share.directory, transfer_params, :overwrite => overwrite_val, :source_base => source_base, :symlink => symlink).perform

      delivery.create_fake_upload(source_share, dest_share, ESession::DIRECTION_RELAY) if symlink
    rescue Node::InvalidTransferSpec
      Utils.parse_recipients(ewf.notify_on_error_names || '').each do |name|
        notified_user = User.find_by_name(name)
        next if notified_user && !notified_user.active?
        BackgroundNotifier.deliver_workgroup_file_relay_notification(delivery, self, e_wg, name, dest_share.node.host, dest_share.directory, file_list_first_10, file_list_all, :error)
      end
    end
  end

  def handle_dropbox_upload_end(delivery)
    wg_delivery = delivery.workgroup_deliveries.first
    membership = EMembership.find_by_e_workgroup_id_and_user_id(
      wg_delivery.recipient.id,
      self.e_package_sender_id
    )
    if membership.present? && membership.dbox_extusr_single_use?
      membership.update_attribute(:dbox_extusr_expired, true)
    end
  end

  def handle_external_submission_end invitation
    if invitation.single_use?
      invitation.expire!
    end
  end

  def deliver_upload_result(delivery)
    result = completed? ? "completed successfully" : "failed to start upload"
    sender = delivery.forward? ? delivery.sender : e_package_sender
    delivery.notifications.on_upload.each do |n|
      unless sender.email == n.email
        BackgroundNotifier.deliver_upload_result_cc delivery, self, n, result
      else
        if sender.external_email_user?
           status_url = EConfiguration.get_full_main_server_address +
             external_submission_status_path(package: delivery.dbox_extusr_sender_pkg_passcode)
        else
          status_url = EConfiguration.get_full_main_server_address + sent_path(delivery, local: sender.force_local_login?)
        end

        BackgroundNotifier.deliver_upload_result delivery, self, sender, result, status_url
      end
    end
  end

  def update_package_status(session)
    case session.status
    when ESession::STATUS_STOPPED
      self.e_package_upload_status = ESession::STATUS_STOPPED
    when ESession::STATUS_ERROR, ESession::STATUS_WILLRETRY
      self.e_package_upload_most_recent_error_time = Time.now
      self.e_package_total_elapsed_seconds         = calc_total_elapsed
      self.e_package_upload_status                 = ESession::STATUS_ERROR


    when ESession::STATUS_TRANSFERRING, ESession::STATUS_PAUSED
      self.e_package_upload_start_time = session.e_session_start_notification_time
      self.e_package_contents_exist    = true
      self.e_package_upload_status     = ESession::STATUS_TRANSFERRING
    when ESession::STATUS_COMPLETE
      self.e_package_upload_start_time      = session.e_session_start_notification_time # in case the transfer went to fast to go into ESession::STATUS_TRANSFERRING state
      self.e_package_upload_completion_time = session.e_session_stop_notification_time
      self.e_package_total_elapsed_seconds  = calc_total_elapsed
      self.e_package_contents_exist = true

      self.e_package_upload_status = ESession::STATUS_COMPLETE
    else
      raise RuntimeError, "Unrecognized session status: #{session.status.inspect}"
    end
  end

  def aspera_env_files
    response = storage_node.api_client.search(inner_directory_path,
        :filter => {:basenames => ['*.aspera-env']})
    if response.success?
      response.directory_items
    else
      []
    end
  end

  def check_ear_encrypted
    unless ear_encrypted?
      # if one or more files are encrypted, it's likely (though not for sure)
      # that the user DID encrypt through Connect, so check ear_encrypted for package
      if e_package_total_files > 0 && aspera_env_files.size > 0
        Utils.log_debug("One or more of the uploaded files are encrypted. User appears to have used encryption-at-rest. Setting package ear_encrypted to true.")
        self.ear_encrypted = true
        save
      end
    end
  end

  def user_downloaded?(user_id)
    access_record = self.e_package_accesses.first(:conditions=>["user_id=?", user_id])
    access_record && access_record.e_package_downloaded
  end

  def user_viewed?(user_id)
    access_record = self.e_package_accesses.first(:conditions=>["user_id=?", user_id])
    access_record && access_record.e_package_viewed
  end

  def upload_transfer_uuid
    e_package_uuid
  end

  def download_transfer_uuid
    SecureRandom.uuid
  end

  def toplevel_items
    items_at('/')
  end
  memoize :toplevel_items

  def toplevel_basenames
    toplevel_items.map(&:basename)
  end
  memoize :toplevel_basenames

  def toplevel_paths
    toplevel_items.map(&:path)
  end
  memoize :toplevel_paths

  def package_status_for_inbox_feed
    upload_status_to_status_map = Hash.new('error').merge(ESession::STATUS_COMPLETE => 'complete',
                                                            ESession::STATUS_TRANSFERRING => 'in-transit',
                                                            ESession::STATUS_ERROR => 'error',
                                                            ESession::STATUS_SUBMITTED => 'in-transit')
    upload_status_to_status_map[self.e_package_upload_status]
  end

  def dropbox_package?
    recipient_users.count      == 0 &&
    recipient_workgroups.count == 1 &&
    recipient_workgroups.first.dropbox?
  end

  def interrupted?
    [ESession::STATUS_STOPPED,
     ESession::STATUS_ERROR].include?(self.e_package_upload_status)
  end


  if Rails.env.development?
    # forcibly mark a package as complete and handle the upload completion,
    # useful if node poller fails to mark a package as completed, but dangerous
    # unless the transfer has actually completed successfully
    def complete!
      return unless inner_directory_exists?
      self.e_package_upload_status = ESession::STATUS_COMPLETE
      self.e_package_contents_exist = true
      self.e_package_upload_completion_time = Time.now
      save!
      handle_upload_end
    end
  end

  def entire?(paths)
    items = paths.map {|path| Utils.expand_path(path, '/').sub(/^\//, '')}
    items.include?("") || (toplevel_basenames - items).empty?
  end

  def self.delete_after_download_policy(enabled, condition)
    if enabled
      condition
    else
      DO_NOTHING
    end
  end

  def self.delete_after_download?(policy)
    policy != DO_NOTHING
  end

  def delete_after_download_policy
    EPackage.delete_after_download_policy(delete_after_download?, delete_after_download_condition)
  end

  def delete_after_download_policy=(policy)
    policy = policy.to_i
    self.delete_after_download = EPackage.delete_after_download?(policy)
    self.delete_after_download_condition = policy
  end

  def recipient_string_for_list
    self.deliveries.first.public_recipients.split(', ').collect{|x| x.include?('(fw') ? x.split('(')[0] + '(fw)' : x}.join(', ')
  end

  def e_package
    self
  end

  def contains_directories?
    source_paths = source_share.node_paths(e_package_paths.map(&:e_uploader_local_path))
    source_paths.any? do  |source_path|
      source_share.node.api_client.files_info(source_path).response_parsed['paths'][0]['type'] == 'directory'
    end
  end

private

  def file_list_first_10
    if @file_list_first_10
      @file_list_first_10
    else
      @file_list_first_10 = ''
      counter = 0
      self.e_package_paths.each do |e_package_path|
        break if counter >= 10
        @file_list_first_10 += e_package_path.e_uploader_local_path.split('/').last + '<br>'
        counter += 1
      end
      @file_list_first_10 += '...<br>' if self.e_package_paths.size > 10
      @file_list_first_10
    end
  end

  def file_list_all
    if @file_list_all
      @file_list_all
    else
      @file_list_all = ''
      self.e_package_paths.each do |e_package_path|
        @file_list_all += e_package_path.e_uploader_local_path.split('/').last + '<br>'
      end
      @file_list_all
    end
  end

  #type can either be :finish or :error
  def send_relay_notifications(sess, type)
    transfer_request = sess.transfer_request
    e_wg = transfer_request.e_workgroup
    dest_share = transfer_request.share
    ewf = e_wg.e_workgroup_forwardings.detect{ |e_wg_f| e_wg_f.share_id == dest_share.id}
    Utils.parse_recipients((type == :error ? ewf.notify_on_error_names : ewf.notify_on_complete_names) || '').each do |name|
      notified_user = User.find_by_name(name)
      next if notified_user && !notified_user.active?
      BackgroundNotifier.deliver_workgroup_file_relay_notification(transfer_request.delivery, self, e_wg, name, dest_share.node.host, dest_share.directory, file_list_first_10, file_list_all, type)
    end
  end

  def calc_total_elapsed
    elapsed = self.upload_sessions.inject(0) {|sum, sess| sum + sess.seconds_elapsed }
    elapsed = 1 if elapsed < 1
    elapsed
  end

  def log_violation(user, message)
    warning_message = "Package access violation: user: #{user.full_name}, package: #{self.e_package_uuid} (#{message})"
    Utils.log_warn(warning_message)
  end

end
