class ESession < ActiveRecord::Base
  STATUS_COMPLETE     = "Completed"
  STATUS_TRANSFERRING = "Transferring"
  STATUS_PAUSED       = "Paused"
  STATUS_ERROR        = "Error"
  STATUS_WILLRETRY    = "Willretry"
  STATUS_STOPPED      = "Stopped"
  STATUS_SUBMITTED    = "Submitted" # if no session message received yet
  STATUS_TIMED_OUT    = "Timed out"
  DIRECTION_UPLOAD    = "Upload"
  DIRECTION_DOWNLOAD  = "Download"
  DIRECTION_CREATE    = "Create"
  DIRECTION_RELAY     = "Relay"

  belongs_to  :e_package
  belongs_to  :initiator, :foreign_key => :initiator_user_id, :class_name => 'User'
  belongs_to  :delivery
  belongs_to  :transfer_request
  belongs_to  :share

  named_scope :uploads,   :conditions => {:direction => DIRECTION_UPLOAD}
  named_scope :downloads, :conditions => {:direction => DIRECTION_DOWNLOAD}
  named_scope :completed, :conditions => {:status => STATUS_COMPLETE}
  named_scope :transferring, :conditions => {:status => STATUS_TRANSFERRING}

  attr_accessible :e_session_start_notification_time,
      :e_session_stop_notification_time,
      :status,
      :central_id,
      :e_package_id,
      :e_package,
      :token_uuid,
      :initiator_user_id,
      :initiator,
      :initiator_name,
      :system_user_name,
      :remote_host,
      :direction,
      :error_code,
      :error_description,
      :bytes_written,
      :bytes_transferred,
      :bytes_pretransfer,
      :files_completed,
      :files_failed,
      :files_pretransfer,
      :seconds_elapsed,
      :partial_package,
      :delivery_id,
      :soap_uri,
      :delivery,
      :share_id,
      :share,
      :transfer_request,
      :local_address

  def ended?
    [STATUS_ERROR, STATUS_COMPLETE, STATUS_STOPPED].include?(status)
  end

  def complete?
    status == STATUS_COMPLETE
  end

  def cancel
    soap_cancel_session(central_id)
  end

  def bits_per_second
    rate = (bytes_transferred * 8) / seconds_elapsed
    (rate.nan? || rate.infinite? || rate < 0) ? nil : rate
  rescue
    nil
  end

  def seconds_remaining
    seconds = ((bytes_pretransfer - bytes_written) * 8 / bits_per_second)
    (seconds.nan? || seconds.infinite? || seconds < 0) ? nil : seconds.to_i
  rescue
    nil
  end

  def progress_percentage
    return 100.0 if complete? # in case the package was too small for bytes_pretransfer to be calculated
    percent = (bytes_written.to_f / bytes_pretransfer) * 100
    (percent.nan? || percent.infinite? || percent < 0) ? 0 : percent
  rescue
    0
  end

  # whether the session is for a full/partial package transfer
  def package_scope
    partial_package? ? "partial" : "full"
  end

  # can be nil if bytes_pretransfer is nil!
  def bytes_total
    [bytes_written, bytes_pretransfer].max
  end

  # can be nil if files_pretransfer is nil!
  def file_count
    [files_completed, files_pretransfer].max
  end

  def start_time
    if e_session_start_notification_time.present?
      e_session_start_notification_time.utc.iso8601
    else
      nil
    end
  end

  def stop_time
    if e_session_stop_notification_time.present?
      e_session_stop_notification_time.utc.iso8601
    else
      nil
    end
  end

  # hash of stats and their default values, add more as necessary
  def self.default_stats
    {
      :bits_per_second     => 0,
      :bytes_total         => 0,
      :bytes_written       => 0,
      :error_description   => nil,
      :files_completed     => 0,
      :initiator_name      => nil,
      :local_address       => '',
      :package_scope       => nil,
      :progress_percentage => 0.0,
      :remote_host         => '',
      :seconds_elapsed     => 0,
      :seconds_remaining   => nil,
      :status              => nil,
      :start_time          => nil,
      :stop_time           => nil,
      :central_id          => nil
    }
  end

  # generate a hash of display-friendly stats that can be used in views
  def stats
    result = {}
    ESession.default_stats.each_key do |stat_name|
      result[stat_name] = send(stat_name)
    end
    result
  end

  private

  def soap_cancel_session(session_id)
    soap = soap_cancel_session_message(session_id)
    storage_node = delivery.default_destination.node

    logger.info "Submitting cancel order to storage node #{storage_node.name}"
    http = Net::HTTP.new(storage_node.host, storage_node.api_port)
    http.use_ssl = storage_node.ssl?
    http.verify_mode = storage_node.verify_ssl? ? OpenSSL::SSL::VERIFY_PEER : OpenSSL::SSL::VERIFY_NONE

    request = Net::HTTP::Post.new('/services/soap/FASPSessionNET-200601/Cancel')
    request.body = soap
    request['Content-Type'] = 'text/xml;charset=UTF-8'
    request['SOAPAction'] = '"FASPSessionNET-200601#Cancel"'
    request.basic_auth(storage_node.api_username, storage_node.unencrypted_api_password)

    http.request(request)
  rescue => e
    logger.error "Unable to cancel soap session #{session_id}: #{e}\n#{e.backtrace.join("\n")}"
  end

  def soap_cancel_session_message(session_id)
%(<?xml version="1.0" encoding="utf-8" ?>
<env:Envelope xmlns:xsd="http://www.w3.org/2001/XMLSchema"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">
  <env:Body>
    <n1:CancelRequest xmlns:n1="urn:Aspera:XML:FASPSessionNET:2006/01:Types">
      <SessionID>#{session_id}</SessionID>
    </n1:CancelRequest>
  </env:Body>
</env:Envelope>)
  end

end
