class Node < ActiveRecord::Base
  class InvalidTransferSpec < StandardError; end

  AZURE_HOST_SUFFIX = '.azure.asperaondemand.com'
  AZURE_STATS_SUFFIX = '-worker.azure.asperaondemand.com'

  # storage types
  STORAGE_WINDOWS_AZURE = "Windows Azure"
  STORAGE_WINDOWS_AZURE_SAS = "Windows Azure SAS"
  STORAGE_DEFAULT = "Default"

  # Override database_validation validation for api_port
  def self.validates_numericality_of *attributes
    if attributes.first == :api_port
      attributes.second[:greater_than] = 0
      attributes.second[:less_than] = 65536
    end
    super
  end
  include DatabaseValidation

  has_many :shares, :dependent => :destroy#, :autosave => true # status is set in the observer which is after a save. Manually saving in NodeObserver

  named_scope :using_azure_storage,
      conditions: ['storage_type = ? OR storage_type = ?', STORAGE_WINDOWS_AZURE, STORAGE_WINDOWS_AZURE_SAS]

  validates_presence_of :name, :host, :api_username, :api_port
  validates_presence_of :api_password, :on => :create

  validates_uniqueness_of :name

  validates_length_of :primary_address,
                      :alternate_address,
                      :alternate_address_client_filter,
                      :alternate_address_server_filter,
                      maximum: 255, allow_blank: true

  with_options :if => :windows_azure? do |windows_azure_form|
    windows_azure_form.validates_presence_of :access_key
    windows_azure_form.validates_presence_of :storage_account
    windows_azure_form.validates_presence_of :blob_container
  end

  with_options :if => :windows_azure_sas? do |windows_azure_sas_form|
    windows_azure_sas_form.validates_presence_of :shared_url
  end

  has_two_way_password :api_password

  before_save :set_should_poll

  attr_accessible :name,
                  :host,
                  :api_port,
                  :api_prefix,
                  :api_username,
                  :storage_type,
                  :storage_account,
                  :access_key,
                  :blob_container,
                  :shared_url,
                  :primary_address,
                  :enable_alternate_address,
                  :alternate_address,
                  :alternate_address_client_filter,
                  :alternate_address_server_filter

  attr_accessible :ssl, :status, :status_at, :status_message, :verify_ssl

  has_many :fasp_nodes, primary_key: 'host', foreign_key: 'node_name'

  # URL used to communicate with host.
  def api_url
    String.new.tap do |url|
      url << "#{protocol}://#{host}"
      url << ":#{api_port}" if api_port.present?
      url << "/#{api_prefix}" if api_prefix.present?
    end
  end

  def protocol
    @protocol ||= ssl? ? 'https' : 'http'
  end

  def api_client
    @azure_url ||= azure_url if uses_azure_storage?
    options = {
      :username => api_username,
      :password => unencrypted_api_password,
      :verify_ssl => verify_ssl,
      :storage_type => storage_type,
    }
    options[:destination_root] = @azure_url if uses_azure_storage?
    @api_client ||= NodeApi::Client.new(api_url, options)
  end

  def browse(path, options = {})
    api_client.browse(path, options)
  end

  def search(path, options = {})
    api_client.search(path, options)
  end

  def test
    client = self.api_client

    status_and_status_message = catch(:error) do

      # ping
      response = client.ping
      if response.error?
        message = response.error['user_message']

        # TODO: shouldn't be doing this based on response text.
        if message =~ /The node does not support this feature/
          message = "Incorrect API Prefix or node not found"
        end

        throw :error, [::App::NOT_PINGABLE, message]
      end

      # info
      response = client.info
      if response.error?
        throw :error, [::App::NOT_INFOABLE, response.error['user_message']]
      end

      # browse
      response = client.browse("/")
      if response.error?
        throw :error, [::App::NOT_BROWSABLE, response.error['user_message']]
      end

      # default case
      [nil, nil]
    end
    self.status         = status_and_status_message.first
    self.status_message = status_and_status_message.last
    self.status_at      = Time.now
  end

  def error?
    status.present?
  end

  def get_transfer_spec(direction, requested_params)
    response = transfer_setup(direction, requested_params)
    if response.success?
      spec = TransferParams.new(response.response_parsed['transfer_specs'].first['transfer_spec'])
      spec.tags["aspera"]["allow_dialogs"] = true if uses_azure_storage?
      spec
    else
      logger.error "Unable to get transfer spec: #{response.user_message}"
      raise InvalidTransferSpec, "Unable to get transfer spec: #{response.user_message}"
      nil
    end
  end

  def initiate_transfer(transfer_order)
    submit_job_soap(transfer_order.to_soap)
  rescue Exception => e
    logger.warn "Error sending SOAP request to #{host}:#{api_port}. #{e.message} - #{e.backtrace}"
  end

  def submit_job_soap(soap)
    logger.info "Submitting transfer order to storage node #{name}"
    http = Net::HTTP.new(host, api_port)
    http.use_ssl = ssl?
    http.verify_mode = verify_ssl? ? OpenSSL::SSL::VERIFY_PEER : OpenSSL::SSL::VERIFY_NONE

    request = Net::HTTP::Post.new('/services/soap/JobNET-200601/Submit')
    request.body = soap
    request['Content-Type'] = 'text/xml;charset=UTF-8'
    request['SOAPAction'] = '"JobNET-200601#Submit"'
    request.basic_auth(api_username, unencrypted_api_password)

    http.request(request)
  rescue Timeout::Error
    logger.warn "Timeout sending SOAP request to #{host}:#{api_port}"
    raise
  end

  def get_token(direction, paths)
    spec = get_transfer_spec(direction, TransferParams.new, paths)
    if spec
      spec.token
    else
      nil
    end
  end

  def host_with_port
    "#{host}:#{api_port}"
  end

  # Address of Central server from user perspective
  def get_server_address http_header_client, http_header_server
    if enable_alternate_address?
      manual_config_server_address_for_headers http_header_client, http_header_server
    else
      fallback_primary_address
    end
  end

  def local?
    host == '127.0.0.1' || host.downcase == FASPEX_CONFIG['Hostname'].downcase
  end

  def api_user_initialized?
    api_username.present? && encrypted_api_password.present?
  end

  def fasp_nodes_error
    fasp_nodes_error? ? 'Error' : 'Active'
  end

  # Nodes that are either inactive or have error state
  def problem_fasp_nodes
    (fasp_nodes.inactive + fasp_nodes.select { |fn| fn.err_desc.present? }).uniq
  end

  def fasp_nodes_error?
    if uses_azure_storage?
      !stats_api_reachable?
    else
      fasp_nodes.empty? || problem_fasp_nodes.length == fasp_nodes.length
    end
  end

  def can_destroy?
    shares.all?(&:can_destroy?)
  end

  def has_default_destination_share?
    !can_destroy?
  end

  def supports_symlinks?
    root_response = api_client.root
    if root_response.error?
      false
    else
      root_response.capabilities.include?('symbolic_links')
    end
  end

  def get_azure_stats_collector_hostname
    # Azure hostnames should be in the form SOMETHING.azure.asperaondemand.com
    # This method may need to change if that ever changes.
    host.split(AZURE_HOST_SUFFIX).first + AZURE_STATS_SUFFIX
  end

  def windows_azure?
    storage_type == STORAGE_WINDOWS_AZURE
  end

  def windows_azure_sas?
    storage_type == STORAGE_WINDOWS_AZURE_SAS
  end

  def uses_azure_storage?
    windows_azure? || windows_azure_sas?
  end

private

  def transfer_setup(direction, requested_params)
    raw_xfer_params = requested_params.attributes.reject {|k,v| v.nil?}
    case direction
    when ESession::DIRECTION_DOWNLOAD
      api_client.download(raw_xfer_params)
    when ESession::DIRECTION_UPLOAD, ESession::DIRECTION_RELAY
      api_client.upload(raw_xfer_params)
    else
      raise ArgumentError, "Direction not understood"
    end
  end

  def manual_config_server_address_for_headers http_header_client, http_header_server
    use_primary = true

    client_filter = alternate_address_client_filter
    client_filter_present = !client_filter.nil? && (client_filter != '')
    server_filter = alternate_address_server_filter
    server_filter_present = !server_filter.nil? && (server_filter != '')

    if client_filter_present || server_filter_present
      asterisk_pattern_regexp = Regexp.new(Regexp.escape(Regexp.escape('*')))
      if client_filter_present
        client_filter_regexp = '^' + Regexp.escape(client_filter) # Create a regexp and escape special chars
        client_filter_regexp.gsub!(asterisk_pattern_regexp,'.*') # Replace escaped asterisks with regexp pattern
        if http_header_client.match(client_filter_regexp)
          use_primary = false
        end
      end
      if server_filter_present
        server_filter_regexp = '^' + Regexp.escape(server_filter)
        server_filter_regexp.gsub!(asterisk_pattern_regexp,'.*')
        if http_header_server.match(server_filter_regexp)
          use_primary = false
        end
      end
    end

    return alternate_address unless use_primary
    fallback_primary_address
  end

  def fallback_primary_address
    primary_address.present? ? primary_address : host
  end

  def azure_url
    if windows_azure_sas? && shared_url =~ /(\w+)\.blob\.core\.windows\.net/
      sas_username = $1
      "azu://" + CGI.escape(sas_username) + ":" + CGI.escape(Base64.encode64(shared_url.split('?').last)) + "@blob.core.windows.net" + get_sas_blob_container
    elsif windows_azure?
      "azu://" + CGI.escape(storage_account) + ":" + CGI.escape(access_key) + "@blob.core.windows.net" + formatted_blob_container_name
    end
  end

  def get_sas_username
    shared_url.scan(/\w+\.blob\.core\.windows\.net/).first.split('.').first
  end

  def get_sas_blob_container
    '/' + shared_url.split('?').first.split('/').last
  end

  def formatted_blob_container_name
    blob_container.first == '/' ? blob_container : '/' + blob_container
  end

  def set_should_poll
    self.should_poll = !uses_azure_storage?
    nil
  end
end
