require 'shellwords'
require 'open3'

class Entitlement
  attr_reader :error_message
  attr_accessor :customer_id, :entitlement_key

  DEFAULT_YEARS_TO_EXPIRATION = 10

  def initialize(customer_id = nil, entitlement_key=nil)
    @customer_id = customer_id
    @entitlement_key = entitlement_key
  end

  def entitled?
    @entitled ||= application_entitled?
    if @entitled && !stale_data?
      return true
    end
    alee_admin status_command
  end

  def register
    alee_admin register_command(customer_id, entitlement_key)
  end

  def unregister
    alee_admin unregister_command
    @customer_id = ''
    @entitlement_key = ''
  end

  def create_internal_user(node_user, node_password)
    result = system(create_internal_user_command(node_user, node_password))

    if !result
      @error_message = 'Unable to add user. Return code #{$?.existstatus}'
      return false
    else
      return true
    end
  end

  def entitle_application
    @e_configuration = EConfiguration.first

    refresh_safe_net_data

    if EConfiguration.safe_net_failed? || !EConfiguration.safe_net_faspex_enabled?
      EConfiguration.first.update_attributes(entitled: false)
      return false
    else
      EConfiguration.first.update_attributes(entitled: true, safe_net_failure_message: nil)
      return true
    end
  end

  def fetch_entitlement_customer_id_and_key
    result_txt, status = Open3.capture2("#{alee_admin_path} entitlement | grep 'ID'")

    return false if status.exitstatus != 0

    customer_id_txt, license_id_txt  = result_txt.split("\n")
    if customer_id_txt.present? &&  license_id_txt.present?
      if customer_id_txt.index(':') && license_id_txt.index(':')
        @customer_id = customer_id_txt.split(':')[1].strip
        @entitlement_key = license_id_txt.split(':')[1].strip
      else
        return false
      end
    else
      return false
    end
  end

  def stale_data?
    if EConfiguration.safe_net_last_updated_at < 3.hours.ago
      EConfiguration.first.update_attribute(:safe_net_failure_message, "Stale Safe Net Data Failure")
      return true
    end
  end

  class << self
    def update_entitlement_license_attributes
      Entitlement.new.entitle_application if EConfiguration.safe_net_entitlement_mode_on?
    end
  end

  private

  def alee_admin(command)
    result = system("#{alee_admin_path} #{command}")

    sleep 2 # This is to compensate (work around) for a node bug where node doesn't update the safenet keys instantly.

    result = system("#{alee_admin_path} ekvget faspex_enabled | grep -q 'faspex_enabled.*yes'")

    if result.nil?
      @error_message = 'Unable to check entitlements'
      return false
    end

    if result
      @entitled = true
      entitle_application
      @error_message = nil
    else
      @error_message = 'Not entitled'
    end

    result
  end

  def status_command
    'status'
  end

  def register_command(customer_id, entitlement_key)
    "register #{customer_id.to_s.shellescape} #{entitlement_key.to_s.shellescape}"
  end

  def unregister_command
    "unregister"
  end

  def alee_admin_path
    '/opt/aspera/bin/alee-admin'
  end

  def application_entitled?
    EConfiguration.first.entitled?
  end

  def api_url
    String.new.tap do |url|
      url << "#{EConfiguration.entitlement_node_use_ssl? ? 'https' : 'http'}://#{EConfiguration.entitlement_node_address}"
      url << ":#{EConfiguration.entitlement_node_port}" if EConfiguration.entitlement_node_port.present?
    end
  end

  def api_client
    @api_client ||= NodeApi::Client.new(api_url,
                                        {
                                          username: EConfiguration.entitlement_node_username,
                                          password: EConfiguration.entitlement_node_password,
                                          verify_ssl: false,
                                          storage_type: Node::STORAGE_DEFAULT,
                                        })
  end

  def connect_to_node?
    begin
      response = api_client.license
    rescue Exception => e
      return false
    end
    return response.http_response.code.to_i == 200
  end

  def refresh_safe_net_data
    response = api_client.license
    license_keys = response.license_keys

    if license_keys.has_key?('failure') && license_keys['failure'].present?
      @e_configuration.safe_net_failed = true
      @e_configuration.safe_net_failure_message = license_keys['failure']
    else
      @e_configuration.safe_net_failed = false
      @e_configuration.license_keys_json = license_keys.to_json
    end

    if license_keys.has_key?('faspex_enabled') && license_keys['faspex_enabled'].downcase == 'yes'
      @e_configuration.safe_net_faspex_enabled = true
    else
      @e_configuration.safe_net_faspex_enabled = false
    end

    if license_keys.has_key?('faspex_max_active_users')
      @e_configuration.safe_net_max_active_users = license_keys['faspex_max_active_users'].to_i
    end

    if license_keys.has_key?('expiration_date')
      @e_configuration.safe_net_expiration_date = Time.parse(license_keys['expiration_date']).utc
    else
      @e_configuration.safe_net_expiration_date = DEFAULT_YEARS_TO_EXPIRATION.years.from_now
    end

    @e_configuration.safe_net_data_valid = true
    @e_configuration.safe_net_last_updated_at = Time.now
    @e_configuration.save(false)
  end

end
