require_dependency 'transfer_params'

module MultiServer

  class RelayDispatcher
    extend Observable
    add_observer(MultiServer::RelayObserver)

    # Action Codes
    DO_NOTHING = "do-nothing"
    UPLOAD_PACKAGE = "upload-package"


    ##
    # Low-Level Post
    ##

    def self.post_http_request_with_exception_handling(host, port, path, data)
      begin
        resp = MultiServer::Synchronizer.send_http_post(host, port, path, data)
        verify_http_response(resp)
        return resp
      rescue SocketError
        http_post_error("no connection to remote server")
      rescue Errno::ECONNREFUSED
        http_post_error("remote server refused connection")
      rescue OpenSSL::SSL::SSLError
        http_post_error("SSL error: #{$!.to_s}")
      rescue Timeout::Error
        http_post_error("Remote server timed out: #{$!.to_s}")
      rescue PackagePostFailed
        raise $!
      rescue
        Rails.logger.warn("$!.inspect: ||#{$!.inspect}||")
        err_msg =
          "Unknown error occurred while posting package to" +
          " server: #{$!.to_s}\n#{$!.backtrace.join("\n")}"
        Rails.logger.debug(err_msg)
        http_post_error("unknown error: #{$!.to_s}")
      end
    end

    def self.verify_http_response(resp)
      valid_resp = (resp.respond_to?(:code) && resp.respond_to?(:body))
      raise RuntimeError, "Invalid response object" if !valid_resp
      case resp.code.to_s
      when '401'
        http_post_error("remote server responded: unauthorized")
      when '503'
        http_post_error("remote server temporarily unavailable")
      when '200'
      else
        http_post_error("remote server responded:\nCode:#{resp.code}\nMessage:#{resp.body}")
      end
    end

    def self.http_post_error(err_msg)
      raise PackagePostFailed.new(err_msg)
    end


    ##
    # Post Relay
    ##

    def self.post_relay(relay, relative_path)
      home_server = relay.relay_home_server
      host = home_server.hostname
      port = home_server.https_port
      path = home_server.url_namespace + relative_path
      data = post_data_for(relay)
      post_http_request_with_exception_handling(host, port, path, data)
    end

    def self.post_data_for(relay)
      return package_post_data_for(relay) if relay.package?
      return access_post_data_for(relay) if relay.access?
      unknown_relay_type(relay)
    end

    def self.package_post_data_for(relay)
      delivery = relay.delivery
      pkg = delivery.e_package
      return {
        :package_to_name => delivery.recipients,
        :package_file_list => pkg.toplevel_basenames,
        :package_name => pkg.e_package_name,
        :package_note => pkg.e_package_note,
        :original_sender_name => pkg.sender.name,
        :package_uuid => pkg.e_package_uuid,
        :metadata_human_readable => pkg.metadata_human_readable,
        :metadata_json => delivery.metadata_json,
        :delivery_uuid => delivery.uuid,
        :delivery_sender_name => delivery.sender_name,
        :delivery_recipients => delivery.public_recipients,
        :delivery_private_recipients => delivery.private_recipients,
        :delivery_notifiable_on_upload => delivery.notifiable_on_upload,
        :delivery_notifiable_on_download => delivery.notifiable_on_download,
        :delivery_title => delivery.title,
        :delivery_note => delivery.note,
        :forward => delivery.forward?,
        :delete_after_download => pkg.delete_after_download?,
        :delete_after_download_condition => pkg.delete_after_download_condition,
        :external_emails => YAML.dump(relay.external_emails)
      }
    end

    def self.access_post_data_for(relay)
      delivery = relay.delivery
      pkg = delivery.e_package
      user = relay.user
      session = relay.e_session

      session_attrs = session.attributes.dup
      session_attrs.delete('delivery_id')
      session_attrs.delete('e_package_id')
      session_attrs.delete('initiator_user_id')

      return {
        :package_uuid => pkg.e_package_uuid,
        :delivery_uuid => delivery.uuid,
        :user_name => user.name,
        :is_partial => relay.is_partial,
        :e_session => YAML.dump(session_attrs)
      }
    end

    def self.parse_post_response(resp)
      Marshal.load(resp.body)
    end


    ##
    # Relay
    ##

    def self.relay_to_remote(relay)
      begin
        relay_to_remote_noex(relay)
      rescue PackagePostFailed, Exception
        relay_failure(relay, $!)
      end
    end

    def self.relay_failure(relay, err)
      begin
        pkg_title = (relay.delivery ? relay.delivery.title.inspect : "unknown-package")
        hostname = (relay.relay_home_server ? relay.relay_home_server.hostname.inspect : "unknown-home-server")
        recipients = (relay.delivery ? relay.delivery.recipients : "unknown-recipients")
        Rails.logger.warn("\nerror relaying package " +
                           "#{pkg_title} " +
                           "to home server #{hostname}" +
                           ", addressed to #{recipients}.\n" +
                           "The error encountered was: #{err.to_s}")
        Rails.logger.debug(err.backtrace.join("\n")) if err.respond_to?(:backtrace)
        relay.increment_failure_count(err.to_s[0...50000])
        relay.update_relay_state(PackageRelay::RETRY) unless relay.relay_state == PackageRelay::FAILED
      rescue MaxRelayFailures
        critical_package_relay_failure(relay, $!) if relay.package?
        critical_access_relay_failure(relay, $!) if relay.access?
      end
    end


    ##
    # Relay
    ##

    def self.relay_to_remote_noex(relay)
      return if relay.relayed?
      log_relay_start(relay)
      path = relative_relay_uri(relay)
      resp = post_relay(relay, path)
      relay_hash = parse_post_response(resp)
      log_relay_response(relay, relay_hash)
      process_relay_response(relay, relay_hash)
    end

    def self.process_relay_response(relay, relay_hash)
      return process_package_relay_response(relay, relay_hash) if relay.package?
      return process_access_relay_response(relay, relay_hash) if relay.access?
      unknown_relay_type(relay)
    end

    def self.log_relay_start(relay)
      return log_package_relay_start(relay) if relay.package?
      return log_access_relay_start(relay) if relay.access?
      unknown_relay_type(relay)
    end

    def self.log_relay_response(relay, relay_hash)
      Rails.logger.debug("remote responded: #{relay_hash.inspect}")
    end


    ##
    # Relay Pacakge
    ##

    def self.process_package_relay_response(relay, relay_hash)
      if relay_hash[:dispatcher_action] == DO_NOTHING
        relay.relay_state = PackageRelay::COMPLETE
        relay.set_last_status("")
        relay.save!
        Rails.logger.debug("no upload necessary, relay complete\n")
      else
        paths = relay.delivery.e_package.toplevel_paths
        transfer_spec = relay_hash[:relay_spec]
        transfer_spec.cookie = TransferCookie.for_transfer(ESession::DIRECTION_RELAY, relay.delivery.sender, relay.delivery)
        transfer_spec.target_rate_cap_kbps = transfer_spec.target_rate_kbps = CConfiguration.first.c_relay_outgoing_bandwidth
        transfer_order = transfer_spec.to_transfer_order(paths)
        transfer_order.retry_count = 0 # will manually retry relays
        delivery.default_destination.node.initiate_transfer(transfer_order)
        relay.update_relay_state(PackageRelay::RELAYING)
        Rails.logger.debug("submitted soap job request, package is uploading\n")
      end
    end

    def self.critical_package_relay_failure(relay, err)
      Rails.logger.warn("relay failed max number of times, marking as failed.\n")
      Rails.logger.debug(err.backtrace.join("\n")) if err.respond_to?(:backtrace)
      relay.update_relay_state(PackageRelay::FAILED)
      type = RelayObserver::MAX_RELAY_FAILURES
      notify_observers_now(:type => type, :relay => relay, :message => err.to_s)
    end

    def self.log_package_relay_start(relay)
      Rails.logger.debug("\n- - - - - - - - -\nrelaying package " +
                          "#{relay.delivery.title.inspect} " +
                          "to home server #{relay.relay_home_server.hostname.inspect}" +
                          ", addressed to #{relay.delivery.recipients}")
    end


    ##
    # Relay Access
    ##

    def self.process_access_relay_response(relay, relay_hash)
      relay.relay_state = PackageRelay::COMPLETE
      relay.set_last_status("")
      relay.save!
      Rails.logger.debug("relay complete\n")
    end

    def self.critical_access_relay_failure(relay, err)
      Rails.logger.warn("package-acess relay failed max number of times, marking as failed.\n")
      Rails.logger.debug(err.backtrace.join("\n")) if err.respond_to?(:backtrace)
      relay.update_relay_state(PackageRelay::FAILED)
    end

    def self.log_access_relay_start(relay)
      Rails.logger.debug("\n- - - - - - - - -\nrelaying package-access for package " +
                          "#{relay.delivery.title.inspect} " +
                          "by user #{relay.user.name.inspect} " +
                          "to home server #{relay.relay_home_server.hostname.inspect}")
    end

    ##
    # Convenience
    ##

    def self.notify_observers_now(event)
      changed
      notify_observers(event)
    end

    def self.soap_cookie(soap_session)
      app_data = soap_session["ApplicationData"]
      cookie_str = (app_data ? app_data["Cookie"] : nil)
      return TransferCookie.new(cookie_str)
    end

    def self.relative_relay_uri(relay)
      return '/package_relay/relay_package' if relay.package?
      return '/package_relay/relay_access' if relay.access?
      unknown_relay_type(relay)
    end

    def self.unknown_relay_type(relay)
      raise RuntimeError, "Relay of unknown type #{relay.relay_type.inspect}"
    end


    ##
    # Background Work
    ##

    def self.next_pending_of_type(relay_type, limit, interval)
      conditions = ["relay_type = ? AND " +
                    "(relay_state = ? OR (relay_state = ? AND last_status_time <= ?))",
                    relay_type,
                    PackageRelay::PENDING,
                    PackageRelay::RETRY,
                    Time.now - interval]
      PackageRelay.find(:all,
                        :conditions => conditions,
                        :include => :delivery,
                        :order => "created_at ASC",
                        :limit => limit)
    end


    def self.next_pending_package_relays
      limit = Config.package_relay_batch_size
      interval = Config.package_relay_retry_interval
      next_pending_of_type(PackageRelay::PACKAGE, limit, interval)
    end

    def self.relay_pending_package_relays
      next_packages = next_pending_package_relays
      next_packages.each do |pkg_relay|
        MultiServer::RelayDispatcher.relay_to_remote(pkg_relay)
      end
    end

    def self.next_pending_access_relays
      limit = Config.access_relay_batch_size
      interval = Config.access_relay_retry_interval
      next_pending_of_type(PackageRelay::ACCESS, limit, interval)
    end

    def self.relay_pending_access_relays
      next_accesses = next_pending_access_relays
      next_accesses.each do |access_relay|
        MultiServer::RelayDispatcher.relay_to_remote(access_relay)
      end
    end

  end
end
