
module MultiServer

  class RelayReceiver


    ##
    # Receive Package Access Relay
    ##

    def self.receive_access_relay(access_data, requesting_ip)
      begin
        receive_access_relay_noex(access_data, requesting_ip)
      rescue RelayAuthenticationFailed
        raise
      rescue RuntimeError
        bt = $!.backtrace.join("\n")
        Rails.logger.warn("\nerror while receiving package-access: #{$!.to_s}")
        Rails.logger.debug(bt)
        access_receive_failure(access_data, $!)
      end
    end

    def self.receive_access_relay_noex(access_data, requesting_ip)
      is_partial = access_data[:is_partial]

      package, delivery, user = parse_access_data(access_data)
      if package.nil? || delivery.nil?
        Rails.logger.debug "Received access info for package with uuid #{access_data[:package_uuid]}, but package does not exist on this server."
        return Marshal.dump({})
      end
      Rails.logger.debug("\n- - - - - - - - \nReceiving package-access " +
          "for delivery with uuid #{delivery.uuid} " +
          "of package with uuid #{package.e_package_uuid} " +
          "downloaded by user #{user.name} ")

      authorize_ip(requesting_ip)
      mark_as_downloaded(package, user)
      create_session_for_access(user, package, delivery, access_data[:e_session])
      delivery.send_del_dled_notif(user, is_partial)
      package.delete_if_download_condition_met
      Marshal.dump({})
    end

    def self.parse_access_data(access_data)
      pkg_uuid = access_data[:package_uuid]
      delivery_uuid = access_data[:delivery_uuid]
      user_name = access_data[:user_name]

      package = EPackage.find_by_e_package_uuid(pkg_uuid)
      delivery = Delivery.find_by_uuid(delivery_uuid)
      user = User.find_by_name(user_name)
      raise RelayFailed, "user #{user_name} does not exist on this server" if user.nil?
      [package, delivery, user]
    end


    ##
    # Receive Package Relay
    ##

    def self.receive_package_relay(relay)
      begin
        receive_package_relay_noex(relay)
      rescue RelayAuthenticationFailed
        raise
      rescue RuntimeError
        bt = $!.backtrace.join("\n")
        Rails.logger.warn("\nerror while receiving package: #{$!.to_s}")
        Rails.logger.debug(bt)
        package_receive_failure(relay, $!)
      end
    end

    def self.receive_package_relay_noex(relay)
      Rails.logger.debug("\n- - - - - - - - \nReceiving package " +
                     "#{relay.package_name.inspect} from " +
                     "#{relay.requesting_ip.inspect}, addressed " +
                     "to #{relay.package_to_name.inspect}.")
      authorize_ip(relay.requesting_ip)
      create_external_users(relay)
      relay.find_sender
      if relay.upload_complete?
        receive_relay_contents_exist(relay)
      elsif relay.package_exists?
        receive_relay_package_exists(relay)
      else
        receive_relay_no_pkg(relay)
      end
    end

    def self.receive_relay_contents_exist(relay)
      Rails.logger.debug("contents exist, no relay necessary.")
      d = find_or_create_delivery(relay)
      d.send_received_notifications(d.default_destination)
      result = relay_complete(relay, RelayDispatcher::DO_NOTHING)
      PostProcess.match_and_execute(d)
      result
    end

    def self.receive_relay_package_exists(relay)
      Rails.logger.debug("package record exists but its " +
                     "contents do not, an upload is necessary.")
      find_or_create_delivery(relay)
      relay_complete(relay, RelayDispatcher::UPLOAD_PACKAGE, relay.transfer_spec)
    end

    def self.receive_relay_no_pkg(relay)
      Rails.logger.debug("creating package record, an upload is necessary.")
      create_delivery(relay)
      relay_complete(relay, RelayDispatcher::UPLOAD_PACKAGE, relay.transfer_spec)
    end

    def self.relay_complete(relay, dispatcher_action, relay_spec = nil)
      relay_hash = {:dispatcher_action => dispatcher_action}
      dispatcher_should_upload = (dispatcher_action == RelayDispatcher::UPLOAD_PACKAGE)
      Rails.logger.debug("relay complete\n") if !dispatcher_should_upload
      Rails.logger.debug("remote dispatch will initiate upload\n") if dispatcher_should_upload
      relay_hash.merge!(upload_response_hash(relay, relay_spec)) if dispatcher_should_upload
      Marshal.dump(relay_hash)
    end

    def self.package_receive_failure(relay, err)
      msg = "Could not process relay: #{err.to_s}. \n" +
        "The relay hash being processed was: #{relay.params.to_yaml}"
      Rails.logger.warn("\n#{msg}")
      Rails.logger.debug(err.backtrace.join("\n"))
      raise RelayFailed, msg
    end

    def self.access_receive_failure(access_data, err)
      msg = "Could not process package-access relay: #{err.to_s}. \n" +
        "The access data being processed was: #{access_data.to_yaml}"
      Rails.logger.warn("\n#{msg}")
      Rails.logger.debug(err.backtrace.join("\n"))
      raise RelayFailed, msg
    end


    ##
    # Resources
    ##

    def self.find_or_create_delivery(relay)
      delivery = create_delivery(relay) if !relay.delivery_exists?
      return relay.delivery
    end

    def self.create_delivery(relay)
      pkg = relay.package
      delivery = Delivery.new(
        :uuid => relay.delivery_uuid,
        :e_package_id => pkg && pkg.id,
        :package_uuid => relay.package_uuid,
        :sender_id => relay.sender.id,
        :sender_name => relay.sender.name,
        :recipients => relay.delivery_recipients || [],
        :private_recipients => relay.delivery_private_recipients || [],
        :notifiable_on_upload => relay.delivery_notifiable_on_upload || [],
        :notifiable_on_download => relay.delivery_notifiable_on_download || [],
        :note => relay.delivery_note,
        :title => relay.delivery_title,
        :forward => relay.forward?,
        :relay => true,
        :metadata_json => relay.metadata_json,
        :skip_path_validation => true,
        :skip_source_validation => true,
        :skip_metadata_validation => true,
        :skip_recipient_validation => true,
        :delete_after_download_policy => relay.delete_after_download_condition
      )
      raise RelayDeliveriesFailed, "Failed to create deliveries" unless delivery.save
      return delivery
    end

    ##
    # Concenience
    ##

    # amos todo: tests
    def self.upload_response_hash(relay, relay_spec)
      c_conf = CConfiguration.find(:first)
      e_conf = EConfiguration.find(:first)
      server_address = CConfiguration.get_server_address(relay.requesting_ip,
                                                         relay.requesting_host)
      {
        :relay_spec => relay_spec,
        :user_name => relay.sender.name,
        :server_address => server_address,
        :path_within_pkg => nil,
        :udp_port => c_conf.c_fasp_udp_port,
        :ssh_port => c_conf.c_ssh_tcp_port,
        :package_home_directory => relay.package.inner_directory_path,
        :remote_incoming_relay_bandwidth => c_conf.c_relay_incoming_bandwidth,
        :package_name => relay.package.e_package_name
      }
    end

    def self.create_external_users(relay)
      relay.external_emails.each do |external_email|
        ExternalUser.find_or_create_by_name(external_email)
      end
    end

    def self.authorize_ip(requesting_ip)
      return if authorized_ip?(requesting_ip)
      raise RelayAuthenticationFailed, "Submitting server is unauthorized for relay"
    end

    def self.mark_as_downloaded(package, user)
      package.mark_as_downloaded_for_user(user.id)
    end

    def self.create_session_for_access(user, package, delivery, encoded_session_parameters)
      session_params = YAML.load(encoded_session_parameters)
      attrs = session_params.merge({
        :delivery => delivery,
        :e_package_id => package.id,
        :initiator_user_id => user.id
      })
      attrs.delete('id')
      session = ESession.new(attrs)
      session.save
      session
    end

    def self.authorized_ips
      ips = []
      HomeServer.all.each do |home_server|
        ips |= [home_server.hostname, home_server.alternate_address1,
                home_server.alternate_address2, home_server.alternate_address3]
      end
      return ips
    end

    def self.authorized_ip?(ip)
      authorized_ips.include?(ip.strip)
    end

  end
end
