# Package methods
#
# This module should be included in any controller that uses packages in its
# view.
#
# Usage:
#   1. require 'packages/package_methods'
#   2. include PackageMethods - adds access methods in class
#   note: helper_method :method - allows view to use this method

require 'systemu'

module PackageMethods

  include ErrorRenderer

  def new
    @dropbox ||= EWorkgroup.find_by_id(params[:dropbox_id]) # nil if not found
    unless logged_in_user_can_send?
      flash[:error] = @error unless @error.blank?
      redirect_to @redirect_path and return
    end
    @form_title = @dropbox ? "Send to Dropbox: #{@dropbox.addressable_name}" : "New Package"
    @metadata = active_metadata_fields
    render :template => 'send/new'
  end

  def check_send_permissions
    unless logged_in_user_can_send?
      respond_to do |format|
        format.html do
          flash[:error] = @error unless @error.blank?
          redirect_to @redirect_path and return
        end
        format.json do
          render_json_error(:unauthorized)
        end
      end
    end
  end

  def prepare_submit
    params[:delivery][:sender] = current_user
    params.delete(:ear_encrypted) unless ear_optional?

    params[:delivery][:delete_after_download_policy] =
        current_user.effective_delete_after_download_policy(params[:delivery][:delete_after_download_policy])

    if @dropbox
      params[:delivery][:recipients] = @dropbox.addressable_name
    elsif params[:passcode]
      invitation = Invitation.find_by_passcode params[:passcode]
      params[:delivery][:recipients] = invitation.inviter.addressable_name
      params[:delivery][:invitation] = invitation
      if invitation.inviter.is_a?(EWorkgroup) && invitation.inviter.dropbox?
        @dropbox = invitation.inviter
      end
    elsif params[:delivery][:recipients_list]
      params[:delivery][:disallow_dropbox] = true
      params[:delivery][:recipients]         = params[:delivery].delete(:recipients_list)
      params[:delivery][:private_recipients] = params[:delivery].delete(:private_recipients_list)
    end

    params[:delivery][:sources] ||= [{id: 1}]
    params[:delivery][:sources].first[:id] ||= 1

    if params[:delivery][:source_paths_list]
      source_paths_list = params[:delivery].delete(:source_paths_list)
      paths = source_paths_list.gsub("\r", "").split("\n") # IE passes in carriage return
      params[:delivery][:sources].first[:paths] = paths
    end

  end

  def prepare_delivery_metadata
    if params[:delivery][:metadata_json]
      metadata_json = params[:delivery].delete(:metadata_json)
      params[:delivery][:metadata] = set_metadata_order(ActiveSupport::JSON.decode(metadata_json.to_s)) || {}
    elsif params[:delivery][:metadata]
      params[:delivery][:metadata] = set_metadata_order params[:delivery][:metadata]
    else
      params[:delivery][:metadata] = {}
    end
  end

  def submit
    prepare_submit

    # #private_recipients= and #public_recipients= require that sender is already set
    @delivery = Delivery.new(sender: current_user)
    @delivery.attributes = params[:delivery]

    #When called from api (ex: /send)  @dropbox is not set as '#new' is not called.
    @dropbox ||= @delivery.dropbox

    prepare_delivery_metadata
    @delivery.metadata = params[:delivery][:metadata]

    if from_remote_source? && @delivery.make_cloud_references? && @delivery.pkg_contains_directories?
      raise EPackage::SendError, "Cannot create references for directories"
    end

    raise EPackage::SendError, formatted_error_messages(@delivery) unless @delivery.save
    @redirect_url   = sent_index_url

    if from_remote_source?
      @delivery.upload_from_source
      render_remote_submit
    else
      for_fasp_url = request.format.json?
      @upload_spec = get_upload_spec @delivery, for_fasp_url
      @upload_spec_hash = JSON.parse(@upload_spec.to_json)
      @upload_spec_hash.reject! {|k, v| v.blank?}
      @upload_spec_hash["authentication"] = "token"
      if @dropbox
        handle_dropbox_submit(@delivery)
      elsif controller_path.starts_with? 'external/submissions'
        @redirect_url = external_submission_status_url package: @delivery.dbox_extusr_sender_pkg_passcode
        BackgroundNotifier.deliver_personal_submit_confirmation @delivery
      end

      if @delivery.http_transfer?
        render_http_submit
      else
        render_connect_submit
      end
    end
  rescue => e
    handle_submit_error(e, @delivery)
  end

  def handle_submit_error(e, delivery)
    if e.is_a?(EPackage::SendError)
      flash.now[:error] = e.to_s
    else
      logger.error("Package creation failed: #{e}\n#{e.backtrace.join("\n")}")
      flash.now[:error] = "Package creation failed."
    end
    respond_to do |format|
      format.js { render 'send/submit_error' }
      format.json do
        render json: package_creation_failed_json(e, delivery)
      end
    end
  end

  def package_creation_failed_json(exception, delivery)
    {
      error: {
        code: delivery.error_code || ErrorCodes::PACKAGE_CREATION_FAILED,
        user_message: 'Package creation failed',
        internal_message: delivery.error_message || exception.to_s
      }
    }
  end

  def render_forward_submit
    respond_to do |format|
      format.js { render template: 'send/forward_submit' }
      format.json{ render json: forward_submit_response }
      format.xml { render xml: forward_submit_response.to_xml(root: 'result') }
    end
  end

  def render_http_submit
    respond_to do |format|
      format.js { render template: 'send/http_submit' }
    end
  end

  def render_remote_submit
    respond_to do |format|
      format.html { redirect_to @redirect_url }
      format.js { render template: 'send/remote_submit' }
      format.json { render json: submit_response }
      format.xml {render xml: submit_response.to_xml(root: 'result')}
    end
  end


  def render_connect_submit
    respond_to do |format|
      format.js { render template: 'send/connect_submit' }
      format.json { render json: submit_response }
      format.xml {render xml: submit_response.to_xml(root: 'result')}
    end
  end

  def forward_submit_response
    response = { links: { status: sent_url(@delivery) } }
  end

  def submit_response
    response = { links: { status: sent_url(@delivery) } }
    if !@delivery.warnings.empty?
      response.merge!(warnings: @delivery.warnings)
    end
    unless from_remote_source?
      response.merge!(xfer_sessions: [@upload_spec.attributes])
    end
    response
  end

  def ajax_retry_package_upload
    @direction = ESession::DIRECTION_UPLOAD
    @display_string = ""
    @ready_to_transfer = false # Initialize flag that informs RJS file whether to transfer or just show an error msg
    error_if_next_step_fails = ""

    # Run the entire thing in a begin/rescue, so that any errors get reportd to the client
    begin
      # Find the package to transfer, show error if it cannot be found
      error_if_next_step_fails = "Error retrieving package"
      pkg_passcode = params[:package]
      if pkg_passcode.present?
        @delivery = Delivery.find_by_dbox_extusr_sender_pkg_passcode(pkg_passcode)
        @package = @delivery.try(:e_package)
      else
        @delivery = Delivery.find(params[:id])
        @package = @delivery.e_package
      end

      error_if_next_step_fails = 'You are not authorized to upload this package'
      if @delivery.nil? || !@package.can_be_uploaded_by?(current_user, request.remote_ip)
        raise SecurityError
      end

      # Create transfer URL
      error_if_next_step_fails = "Error while creating transfer URL"
      upload_spec = get_upload_spec(@delivery)
      @upload_spec_hash = JSON.parse(upload_spec.to_json)
      @upload_spec_hash.reject! {|k, v| v.blank?}
      @upload_spec_hash["authentication"] = "token"
      @ready_to_transfer = true #set flag to inform RJS file that we're going ahead with transfer rather than showing an error

      @package.update_attribute(:e_package_upload_status, ESession::STATUS_SUBMITTED)
      # Done
      render :partial => 'packages/ajax_retry_upload'
    rescue
      Utils.log_warn("#{$!.class.to_s}: ajax_retry_package_upload failed: " + error_if_next_step_fails.to_s + ": " + $!.to_s + $!.backtrace.join("\n"))
      # An error occurred
      flash[:error] = error_if_next_step_fails
      render :partial => 'packages/ajax_retry_upload' and return false
    end
  end


  def ajax_stop_upload
    package = params[:package].present? ? @delivery.e_package : Delivery.find(params[:id]).e_package
    unless package.sent_by?(current_user)
      raise SecurityError, "User #{current_user.full_name} tried to cancel another user's transfer"
    end

    package.upload_sessions.last.cancel
    package.update_attribute(:e_package_upload_status, ESession::STATUS_ERROR)
    render :nothing => true
  rescue
    Utils.log_warn("Error trying to cancel transfer: #{$!.to_s}")
  end

  def delivery_id
    user_delivery = UserDelivery.find_by_id(params[:id])
    if user_delivery
      user_delivery.delivery.id
    else
      nil
    end
  end

  def delivery
    Delivery.find(delivery_id)
  end

  # prepares for ajax updating of file browser in package view
  def ajax_update_file_browser
    prep_file_browser
    respond_to do |format|
      format.js { render template: 'packages/browse' }
    end
  end

  def ajax_retry_all_relays
    wd = WorkgroupDelivery.find_by_id(params[:id])
    wd.most_recent_file_relay_requests_by_share.each do |tr|
      if [ESession::STATUS_STOPPED, ESession::STATUS_ERROR, ESession::STATUS_TIMED_OUT].include?(tr.status)
        #Some share could be nil if deleted after a relay to it has take place in the past.
        #So only retry relay on shares which are !nil.
        if share = tr.share
          e_wg = tr.e_workgroup
          ewf = EWorkgroupForwarding.first(:conditions => { :share_id => share.id, :e_workgroup_id => e_wg.id } )
          wd.e_package.relay_to_forwarding_destination(share, tr.delivery, e_wg, ewf)
        end
      end
    end

    @deliveries = [wd]
    @force_icon = true
    @force_status = true
    render :partial => 'packages/ajax_update_wg_packages_list'
  end

  def ajax_update_wg_packages_list
    #get the list of record ids (fed in thru URL)
    id_list_string = params[:record_id_list].split(",").collect{|anid| anid.to_i}.compact.join(",") # This is SQL safe
    @deliveries = []
    unless id_list_string.nil? || id_list_string == ''
      @deliveries = WorkgroupDelivery.all(:conditions => ["id IN (#{id_list_string})"])
    end
    render :partial => 'packages/ajax_update_wg_packages_list'
  end

  def ajax_update_package_list
    id_list_string = params[:record_id_list]
    @records = []
    unless id_list_string.nil? || id_list_string == ''
      @records = EPackage.find(:all, :conditions => ['id IN (?)', id_list_string])
    end
    render :partial => 'packages/ajax_update_package_list'
  end

  def ajax_update_delivery_list
    id_list_string = params[:record_id_list]
    @deliveries = []
    unless id_list_string.nil? || id_list_string == ''
      @deliveries = Delivery.find(:all, :conditions => ['id IN (?)', id_list_string])
    end
    render :partial => 'packages/ajax_update_delivery_list'
  end

  def ajax_update_received_list
    #get the list of record ids (fed in thru URL)
    id_list_string = params[:record_id_list].split(",").collect{|anid| anid.to_i}.compact.join(",") # This is SQL safe
    @deliveries = []
    unless id_list_string.nil? || id_list_string == ''
      @deliveries = UserDelivery.all(:conditions => ["id IN (#{id_list_string.to_s})"])
    end
    render :partial => 'packages/ajax_update_received_list'
  end

  def ajax_update_sent_list
    #get the list of record ids (fed in thru URL)
    id_list_string = params[:record_id_list].split(",").collect{|anid| anid.to_i}.compact.join(",") # This is SQL safe
    @deliveries = []
    unless id_list_string.blank? || id_list_string == ''
      @deliveries = Delivery.all(
        :conditions => ["id IN (#{id_list_string.to_s})"],
        :include => :e_package
      )
    end
    render :partial => 'packages/ajax_update_sent_list'
  end

  # prepares for ajax updating of stats in package view
  def ajax_update_package
    prep_file_browser
    @delivery = delivery
    @package = @delivery.e_package
    @display_stats_hash = @package.upload_statistics
    if (@package.e_package_upload_status == ESession::STATUS_COMPLETE)
      @completed_downloads_by_user = @package.get_completed_downloads_by_user
      @total_completed_downloads = 0
      @completed_downloads_by_user.each {|x| @total_completed_downloads += x[:download_count] }

      @active_downloads = @package.get_active_download_stats
      @completed_partial_downloads_by_user = @package.get_completed_partial_downloads_by_user
      @total_completed_partial_downloads = 0
      @completed_partial_downloads_by_user.each {|x| @total_completed_partial_downloads += x[:partial_download_count]}
    elsif [ESession::STATUS_TRANSFERRING, ESession::STATUS_SUBMITTED].include?(@package.e_package_upload_status)
      @show_spinners = true
    elsif @package.interrupted?
      @hide_spinners = true
    end
    @package_deleted = (!@package.e_package_contents_exist && @package.e_package_status != EPackage::STATUS_VALID && (params[:showing_existant_package].to_i == 1))
    render :partial => 'packages/ajax_update_package'
  end

  def forward_submit
    prepare_submit
    forward_package
    render_forward_submit
  rescue => e
    handle_submit_error e, @delivery
  end

  def find_delivery
    case self.controller_path
    when 'admin/packages'
      Delivery.find(params[:id])
    when 'workgroup/deliveries', 'dropbox/deliveries'
      find_workgroup_delivery(params[:id])
    when 'sent'
      Delivery.find(params[:id])
    else #received list and else
      UserDelivery.find(params[:id])
    end
  end

  def delivery_send
    delivery
  end

  # prepare to show package view
  def show
    @back_url = url_for(:action => params[:view] ? 'history' : 'index') if self.controller_name != 'external_delivery'
    @return_str = params[:view] ? "Return to History" : "Return to List"
    @delivery = delivery
    @package = @delivery.e_package

    # mark package as read
    @package.mark_as_viewed_for_user(current_user.id)
    if @package.e_package_contents_exist? && @package.e_package_status == EPackage::STATUS_VALID
      @showing_existant_package = 1
    else
      @showing_existant_package = 0
    end

    @display_stats_hash = @package.upload_statistics
    if @package.e_package_contents_exist? && downloadable_object.inner_directory_exists? && @package.e_package_status == EPackage::STATUS_VALID
      prep_file_browser
    end

    if (@package.upload_status == ESession::STATUS_COMPLETE)
      @completed_downloads_by_user = @package.get_completed_downloads_by_user
      @active_downloads = @package.get_active_download_stats
      @total_completed_downloads = 0
      @completed_downloads_by_user.each {|x| @total_completed_downloads += x[:download_count] }
      @completed_partial_downloads_by_user = @package.get_completed_partial_downloads_by_user
      @total_completed_partial_downloads = 0
      @completed_partial_downloads_by_user.each {|x| @total_completed_partial_downloads += x[:partial_download_count]}
    end

    respond_to do |format|
      format.html
      format.xml
    end
  end

private

  def formatted_error_messages(delivery)
    error_messages = delivery.errors.full_messages +
        delivery.e_package.errors.full_messages
    error_messages.join("\n")
  end

  def handle_dropbox_submit(delivery)
    wg_delivery = delivery.workgroup_deliveries.first
    if current_user.external_email_user?
      passcode = delivery.dbox_extusr_sender_pkg_passcode
      @redirect_url = external_submission_status_url package: passcode
      mailing_url = EConfiguration.get_full_main_server_address +
        external_submission_status_path(:package => passcode,
                                        :skip_relative_url_root => true)
      BackgroundNotifier.deliver_dropbox_submit_confirmation(wg_delivery, mailing_url)
    end
  end

  def from_remote_source?
    SourceShare.find(params[:delivery][:sources].first[:id]).remote?
  rescue
    false
  end

  def ear_optional?
    @dropbox ? @dropbox.ear_optional? : EConfiguration.ear_optional?
  end

  # checks that the logged in user can send normal packages from ESendController or to a valid, permitted dropbox from DropboxPackagesController
  def logged_in_user_can_send?
    user = current_user
    if user.nil?
      @redirect_path = page_not_found_path
      false
    elsif in_dropbox_deliveries_controller?
      user_can_send_package?(user) && @dropbox && user_can_send_to_dropbox?(user)
    else
      user_can_send_package?(user)
    end
  end

  def user_can_send_package?(user)
    if !user.uploads_allowed && @dropbox.nil?
      @error = "You do not have permission to upload files."
    elsif !user.upload_from_ip_allowed?(request.remote_ip)
      @error = "You do not have permission to upload files from your IP address."
    else
      return true
    end
    @redirect_path = received_index_path
    false
  end

  def user_can_send_to_dropbox?(user)
    membership = EMembership.first(:conditions => ['e_workgroup_id = ? AND user_id  = ?', @dropbox.id, user.id])
    if !user.external_email_user?
      membership.dbox_extusr_passcode = nil
      membership.dbox_extusr_single_use = 0
      membership.dbox_extusr_expires_at = nil
      membership.dbox_extusr_expired = false
      membership.save if membership.changed?
    end

    if membership.nil? || membership.dbox_extusr_passcode != params[:passcode]
      @error = nil # suppress error message
    elsif membership.expired?
      @error = "Your invitation to dropbox #{ERB::Util.h @dropbox.addressable_name} has expired."
    else
      @passcode  = membership.dbox_extusr_passcode
      return true
    end
    @redirect_path = page_not_found_path
    false
  end

  def active_metadata_profile
    if @dropbox
      MetadataProfile.find_by_id(@dropbox.dbox_mdata_profile_id)
    else
      MetadataProfile.find_by_id(EConfiguration.first.normal_pkg_mdata_profile_id)
    end
  end

  # returns nil if there is no active metadata profile
  def active_metadata_profile_id
    profile = active_metadata_profile
    profile ? profile.id : nil
  end

  def active_metadata_fields
    profile = active_metadata_profile
    profile ? profile.all_fields_array : [ ]
  end

  def in_dropbox_deliveries_controller?
    self.class == Dropbox::DeliveriesController
  end

  def find_workgroup_delivery(wg_delivery_id)
    # to keep pre-3.0.3 email links working, we need to look for both id and old id
    wg_delivery = WorkgroupDelivery.find(:first, :conditions => ['id=?',wg_delivery_id]) ||
        WorkgroupDelivery.find(:first, :conditions => ['old_wg_delivery_id=?', wg_delivery_id])

    if wg_delivery.nil?
      raise ActiveRecord::RecordNotFound, "couldn't find WorkgroupDelivery with id= or old_wg_delivery_id= #{wg_delivery_id}"
    else
      wg_delivery
    end
  end


  def check_package_permissions
    #only applies to showing a single package
    return if !["show", "forward"].include?(params[:action])

    logged_in_user = current_user

    if params[:controller] == 'admin/packages'
      requested_package = EPackage.find(Delivery.find(params[:id]).e_package_id)
    elsif params[:controller] == 'workgroup/deliveries' || params[:controller] == 'dropbox/deliveries'
      requested_package = EPackage.find(find_workgroup_delivery(params[:id]).e_package_id)
    elsif params[:controller] == 'sent'
      delivery = Delivery.find(params[:id])
      if delivery.sender_id != logged_in_user.id
        render_unauthorized_package_view_error
        return false #abort rest of the filter chain
      end
      requested_package = EPackage.find(delivery.e_package_id)
    else
      user_delivery = UserDelivery.find(params[:id])
      if user_delivery.recipient_id != logged_in_user.id
        render_unauthorized_package_view_error
        return false #abort rest of filter chain
      end
      requested_package = EPackage.find(user_delivery.e_package_id)
    end

    if !requested_package.accessible_to_user(logged_in_user.id, params[:controller] == "workgroup/deliveries")
      render_unauthorized_package_view_error
      return false #abort rest of filter chain
    end
  end

  def render_unauthorized_package_view_error
    logger.info("INFO: Attempt by unauthorized user (id = " +
                "#{session[:login_user_id]}) to access package (id = #{params[:id]})")
    error_message = "You are unauthorized to view this package."
    render_error(:unauthorized, error_message)
  end

  # prepares for display of file browser
  def prep_file_browser
    @delivery = delivery
    @package = @delivery.e_package
    @relative_path = Utils.sanitize_path(params[:new_path] || '')
    @items = downloadable_object.items_at(@relative_path)
  end

  def get_upload_spec(delivery_send, for_fasp_url = false)
    delivery_send.get_transfer_spec(ESession::DIRECTION_UPLOAD, ['/'], current_user,
        :token =>true, :host => delivery_send.default_transfer_host(request, true), for_fasp_url: for_fasp_url)
  rescue EPackage::SendError
    raise
  rescue => e
    logger.error("Unable to generate fasp URL: #{e}\n#{e.backtrace.join("\n")}")
    raise EPackage::SendError, "Package creation failed"
  end

  def forward_package
    package = params[:package_id] ?  EPackage.find(params[:package_id]) : Delivery.find(params[:id]).e_package
    check_forward_permissions(package)
    @delivery = Delivery.new(sender: current_user)
    @delivery.attributes = params[:delivery].merge(
      e_package: package, skip_path_validation: true, forward: true
    )

    prepare_delivery_metadata
    @delivery.metadata = params[:delivery][:metadata]

    raise EPackage::SendError, formatted_error_messages(@delivery) unless @delivery.save
    @delivery.reload # refresh associations
    PostProcess.match_and_execute(@delivery)
    @delivery.relay_to_inboxes
    Thread.new do
      # Make sure to ignore original delivery when determining share, because we want to send
      # notifications specifically for the forwarded share
      @delivery.send_received_notifications(@delivery.default_destination(true))
    end
    package.create_relays_for(@delivery)
    package.relay_to_forwarding_destinations(@delivery)
  end

  def check_forward_permissions(package)
    unless package.can_be_forwarded_by?(current_user)
      raise EPackage::SendError, "You are not authorized to forward this package"
    end
  end

  def set_metadata_order metadata
    #Hashes enumerate their values in the order that the corresponding keys were inserted.
    inorder_metadata = {}
    metadata_fields = active_metadata_fields

    metadata_fields.each do |field|
      field_name = field[:id_name].to_s
      inorder_metadata[field_name] = metadata[field_name]
    end

    inorder_metadata
  end

end
