# Methods for setting up transfers go here, including generating fasp URLs,
# transfer tokens, and cookies

# including class must define e_package, delivery, and storage_share
module UrlGeneration

  # Options:
  # 'token' - whether or not to provide a token in the transfer spec, defaults to false
  # 'host' - alternate hostname to use in place of host returned by node
  # 'for_fasp_url' - whether this transfer spec is to be used for generating a fasp URL
  def get_transfer_spec(direction, paths_in_pkg, requester, options = {})
    target_share = direction == ESession::DIRECTION_DOWNLOAD ? default_download_location : default_destination
    transfer_settings_owner = target_share.override_transfer_parameters ? target_share : requester
    params = requested_params(direction, paths_in_pkg, requester, transfer_settings_owner, target_share, options[:for_fasp_url].present?)
    spec = target_share.node.get_transfer_spec(direction, params)
    raise 'Unable to get transfer spec from storage node' if spec.nil?
    spec.remote_host = options[:host]
    spec.token = nil unless options[:token]
    spec
  end

  def prepare_transfer(direction, paths_in_pkg, requesting_user, options = {} )
    transfer_cookie = TransferCookie.for_transfer(direction, requesting_user, delivery)
    TransferRequest.create(
      user: requesting_user,
      delivery: delivery,
      e_package: e_package,
      share: options.delete(:share) || storage_share,
      cookie: transfer_cookie.to_s,
      uuid: transfer_cookie.transfer_uuid,
      direction: direction,
      partial_package: direction == ESession::DIRECTION_DOWNLOAD ? !e_package.entire?(paths_in_pkg) : false,
      flatten: options.delete(:flatten) || false,
      e_workgroup_id: options.delete(:e_workgroup_id)
    )
  end

 #private

  def select_transfer_rate_settings(direction, transfer_settings_owner, requester)
    transfer_rate_settings = { max_rate: requester.max_rate(direction), target_rate: requester.target_rate(direction), lock_minrate_policy: requester.lock_minrate_policy }
    if (requester != transfer_settings_owner) && transfer_settings_owner.override_transfer_parameters
      transfer_rate_settings = { target_rate:  [transfer_settings_owner.target_rate(direction), requester.target_rate(direction)].min,
                                 max_rate:  [transfer_settings_owner.max_rate(direction), requester.max_rate(direction)].min,
                                 lock_minrate_policy: (transfer_settings_owner.lock_minrate_policy | requester.lock_minrate_policy) }
    end
    transfer_rate_settings
  end

  def node_paths(paths_in_pkg)
    paths_in_pkg.map { |path| e_package.node_path(path) }
  end

  def create_dir?(direction, paths_in_pkg)
    direction != ESession::DIRECTION_DOWNLOAD || paths_in_pkg.count > 1
  end

  def protect?(direction)
    if direction == ESession::DIRECTION_UPLOAD
      require_ear? || e_package.ear_encrypted?
    else
      e_package.ear_encrypted?
    end
  end

  def content_protection(direction)
    case
    when direction == ESession::DIRECTION_UPLOAD && (require_ear? || e_package.ear_encrypted?)
      "encrypt"
    when direction == ESession::DIRECTION_DOWNLOAD && e_package.ear_encrypted?
      "decrypt"
    else
      nil
    end
  end

  def require_ear?
    if dropbox_delivery?
      dropbox.require_ear?
    else
      EConfiguration.require_ear?
    end
  end

  def requested_params(direction, paths_in_pkg, requester, transfer_settings_owner, target_share, for_fasp_url = false)
    conf = CConfiguration.first
    node = target_share.node
    transfer_rate_settings = select_transfer_rate_settings(direction, transfer_settings_owner, requester)

    params = {
      "cookie" => prepare_transfer(direction, paths_in_pkg, requester).cookie,
      "target_rate_kbps" => transfer_rate_settings[:target_rate],
      "cipher_allowed" => conf.cipher,
      "content_protection" => content_protection(direction),
      "http_fallback" => conf.http_fallback_enabled?,
      "create_dir" => create_dir?(direction, paths_in_pkg),
      "lock_rate_policy" => transfer_rate_settings[:lock_minrate_policy],
      "lock_min_rate" => transfer_rate_settings[:lock_minrate_policy],
      "target_rate_cap_kbps" => transfer_rate_settings[:max_rate],
      "tags" => {
        "aspera" => {
          "faspex" => {
            "requestor" => user_params(requester)
          }
        }
      },
    }.merge!(path_and_root_params(direction, paths_in_pkg, for_fasp_url))

    # if transfer server is local, use our own http/https port instead of the one the
    # node daemon thinks is the right port
    if node.local? || FASPEX_CONFIG['UseApachePortsForHttpFallback']
      params["http_fallback_port"] = EConfiguration.first.e_http_port.to_i
      params["https_fallback_port"] = EConfiguration.first.e_https_port.to_i
    end

    TransferParams.new(params)
  end

  def path_and_root_params(direction, paths_in_pkg, exclude_sources)
    if direction == ESession::DIRECTION_DOWNLOAD
      sources = paths_in_pkg.map do |path_in_pkg|
        { 'source' => download_path(path_in_pkg) }
      end
      {'paths' => sources}
    elsif exclude_sources
      pkg_path = node_path(paths_in_pkg.first)
      destination_root = storage_share.node.api_client.destination_root || '/'

      pkg_path.gsub!(/^\//, '').split('/').inject('') do |path, directory|
        storage_share.api_client.create_directory(File.join(path, directory))
        path = File.join(path, directory)
      end

      {'destination_root' => File.join(destination_root, (storage_share.uses_azure_storage? ? URI.encode(pkg_path) : pkg_path))}
    else
      source_paths = e_package.e_package_paths.map(&:e_uploader_local_path)
      pairs = source_paths.map do |source_path|
        {
          'source' => source_path,
          'destination' => node_path(File.basename(source_path.gsub(/\\/,'/')))
        }
      end
      {'paths' => pairs}
    end
  end

  def user_params(user)
    params = {
      'name' => user.name,
      'first_name' => user.first_name,
      'last_name' => user.last_name,
      'email' => user.email
    }
    user.user_profile_data.each do |attribute|
      params[attribute.user_profile_field.name] = attribute.value
    end
    params
  end

end
