require_dependency 'transfer_cookie'

module UrlRegeneration
  class Regenerator

    def self.issue_token_for_urls(xml, requesting_user_id, requesting_ip, requested_host)
      begin
        urls = parse(xml)
        token = refresh(urls, requesting_user_id, requesting_ip, requested_host)
        return token
      rescue URI::Error
        regeneration_failure(xml, "Could not parse URL: #{$!.to_s}", "invalid-url")
      rescue InvalidTransferCookie
        regeneration_failure(xml, "Url contains invalid cookie: #{$!.to_s}", "invalid-url")
      rescue ActiveRecord::RecordNotFound
        regeneration_failure(xml, $!.to_s, "url-not-found")
      rescue Errors::RegenerationError
        raise $!
      rescue Exception
        Utils.log_info("Error regenerating URLs: #{$!.to_s}")
        Utils.log_debug("#{$!.backtrace.join("\n")}")
        regeneration_failure(xml, $!.to_s, "unknown-error")
      end
    end

    def self.regeneration_failure(xml, err, cond)
      raise Errors::RegenerationError.new(err.to_s, cond)
    end

    def self.parse(raw_xml)
      xml = Nokogiri::XML(raw_xml)
      url_list = xml/'url-list'
      validate_list_element(url_list)
      url_elems = url_list/'url'
      url_elems.collect{|url| parse_url_element(url)}
    end

    def self.validate_list_element(url_list)
      raise Errors::RegenerationError.new("Expected at least one 'url-list' element",
                                          "xml-not-well-formed") if (url_list.length == 0)
      raise Errors::RegenerationError.new("Expected at most one 'url-list' element",
                                          "xml-not-well-formed") if (url_list.length > 1)
    end

    def self.parse_url_element(url)
      validate_url_element(url)
      href = url['href'].to_s
      unless href =~ %r{^(fasp|faspe)://([^/]*)@([^/]*):(\d*)(/.*)\?.*cookie=([^&]*).*}
        raise Errors::RegenerationError.new("Invalid url, unable to parse fasp_url", "invalid_url")
      end
      user, host, port, path, cookie = $2, $3, $4, $5, $6
      params = TransferParams.new(
        remote_user: user,
        remote_host: host,
        ssh_port: port,
        cookie: cookie
      )
      FaspUrl.new(path, params)
    end

    def self.validate_url_element(url)
      raise Errors::RegenerationError.new("Invalid url, missing 'href' attribute: #{url.to_s}",
                                          "invalid-url") if url['href'].nil?
    end

    def self.parse_outer_package_directory_from_url(url)
      url_path = URI.decode(url.path)
      # Get the outer package name -- (docroot/)outer_package_name.aspera-package/inner_package_name
      # Should always be second from last item
      url_path.split('/')[-2]
    end

    def self.refresh(urls, requesting_user_id, requesting_ip, requested_host)
      raise ArgumentError, "issue-token request must contain at least one fasp url" if ( urls.class != Array or urls.length == 0 )

      directory_name = parse_outer_package_directory_from_url(urls[0])

      # verify that all urls being refreshed belong to the same package and are for download
      urls.map { |url|
        raise Errors::RegenerationError.new("Unable to generate token for urls from different packages",
                                            "invalid-url") if parse_outer_package_directory_from_url(url) != directory_name
        cookie = TransferCookie.new(url.transfer_params.cookie)
        raise Errors::RegenerationError.new("Invalid url, url is not for download", "invalid-url") if !cookie.is_download?
      }

      # verify that the package still exists and the user is authorized to download it
      package = EPackage.find_by_outer_directory_name_and_e_package_contents_exist(directory_name, true)

      raise ActiveRecord::RecordNotFound, "URL references package that no longer " +
        "exists: '#{urls[0].to_s}'" if package.nil?
      user = User.find(requesting_user_id)
      raise ActiveRecord::RecordNotFound, "User not found" if user.nil?
      raise SecurityError, "User is not authorized to download this package" if !package.can_be_downloaded_by?(user, requesting_ip)

      return refresh_download(urls, package, requesting_user_id, requesting_ip, requested_host)
    end

    def self.refresh_download(urls, package, requesting_user_id, requesting_ip, requested_host)
      relative_paths = []
      urls.each do |url|
        path = URI.decode(url.path)
        relative_paths << path_relative_to_package_root(package, path)
      end
      requesting_user = User.find_by_id(requesting_user_id)
      spec = package.deliveries.first.get_transfer_spec(
        ESession::DIRECTION_DOWNLOAD, relative_paths, requesting_user, token: true
      )
      spec.token
    end

    private

    def self.path_relative_to_package_root(package, path)
      package_root = Pathname.new(package.inner_directory_path(true))
      Pathname.new(path).relative_path_from(package_root).to_s
    end
  end
end
